(*
 * Usefull functions for building and working with Mizar parse tree.
 * Date: 2006-10-15
 * Copyright (C) 2006 Michael Nedzelsky (MichaelNedzelsky@yandex.ru)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)

open Lexer

type ast_miz = Article | Environment_declaration | Text_proper | Directive | Vocabulary_directive
  | Library_directive of lib_directive | Requirement_directive | Vocabulary_name | Article_name | Requirement | File_name
  | Section | Text_item | Reservation | Definitional_item | Registration_item | Notation_item | Theorem
  | Scheme_item | Auxiliary_item | Canceled_theorem
  | Label_identifier | Theorem_number | Definition_number | Local_reference | Library_reference | Reference
  | References | Scheme_identifier | Local_scheme_reference | Scheme_number | Library_scheme_reference | Scheme_reference
  | Straightforward_justification | Scheme_justification | Simple_justification | Justification | Proof
  | Definitional_block | Definition_item | Definition | Redefinition_block
  | Registration_block | Notation_block | Structure_definition | Mode_definition | Functor_definition
  | Predicate_definition | Attribute_definition | Canceled_definition
  | Mode_pattern | Mode_symbol | Loci | Locus | Variable_identifer
  | Loci_declaration | Permissive_assumption
  | Qualified_variables | Conditions | Proposition | Sentence | Formula_expression
  | Assumption | Single_assumption | Collective_assumption | Existential_assumption
  | Statement | Private_definition
  | Constant_definition | Private_functor_definition | Private_predicate_definition
  | Equating_list | Equating | Term_expression | Private_functor_pattern | Private_predicate_pattern
  | Functor_identifier | Predicate_identifier | Type_expression_list
  | Linkable_statement | Diffuse_statement
  | Compact_statement | Choice_statement | Type_changing_statement | Iterative_equality
  | Reasoning | Reasoning_item | Case_list | Suppose_list | Case | Suppose
  | Skeleton_item | Generalization | Conclusion | Exemplification
  | Diffuse_conclusion | Ancestors | Structure_symbol | Fields
  | Specification | Definiens of ast_miz | Simple_definiens of ast_miz | Conditional_definiens of ast_miz | Partial_definiens_list of ast_miz
  | Partial_definiens of ast_miz
  | Correctness_conditions | Type_expression | Correctness_condition
  | Functor_pattern | Functor_loci | Functor_symbol | Left_functor_bracket | Right_functor_bracket
  | Functor_property | Predicate_property | Predicate_symbol | Attribute_symbol | Attribute_pattern
  | Numeral
  | Cluster_registration | Existential_registration | Conditional_registration | Functorial_registration
  | Adjective | Adjective_cluster | Canceled_registration
  | Notation_declaration | Attribute_synonym | Attribute_antonym | Functor_synonym | Mode_synonym | Predicate_synonym | Predicate_antonym
  | Predicate_pattern | Scheme_block | Scheme_parameters | Scheme_conclusion | Scheme_premise
  | Field_segment | Selector_symbol | Structure_type_expression | Term_expression_list | Type_change_list

(*
let ast_miz_xml =
  [("Arcticle", Article); ("Environment_declaration", Environment_declaration); ("Text_proper", Text_proper);
   ("Directive", Directive); ("Vocabulary_directive", Vocabulary_directive); ("Library_directive", Library_directive);
   ("Requirement_directive", Requirement_directive); ("Vocabulary_name", Vocabulary_name); ("Article_name", Article_name);
   ("Requirement", Requirement); ("File_name", File_name)]

let ast_map = add_list MapOfString.empty ast_miz_xml
*)

(** Returns a pair (tag_name, attributes as association list) *)
let ast_miz_to_tag t =
  let lib_to_attr lib =
    match lib with
    | Notations -> "notations"
    | Constructors -> "constructors"
    | Registrations -> "registrations"
    | Definitions -> "definitions"
    | Theorems -> "theorems"
    | Schemes -> "schemes"
  in
  match t with
(* 1 *)  | Adjective -> ("Adjective", [])
(* 2 *)  | Adjective_cluster -> ("Adjective_cluster", [])
(* 3 *)  | Ancestors -> ("Ancestors", [])
(* 4 *)  (* | Arguments -> ("Arguments", []) *)
(* 5 *)  | Article -> ("Article", [])
(* 6 *)  | Article_name -> ("Article_name", [])
(* 7 *)  | Assumption -> ("Assumption", [])
(* 8 *)  (* | Atomic_formula_expression -> ("Atomic_formula_expression", []) *)
(* 9 *)  | Attribute_antonym -> ("Attribute_antonym", [])
(* 10 *) | Attribute_definition -> ("Attribute_definition", [])
(* 11 *) | Attribute_pattern -> ("Attribute_pattern", [])
(* 12 *) | Attribute_symbol -> ("Attribute_symbol", [])
(* 13 *) | Attribute_synonym -> ("Attribute_synonym", [])
(* 14 *) | Auxiliary_item -> ("Auxiliary_item", [])
(* 15 *) | Canceled_definition -> ("Canceled_definition", [])
(* 16 *) | Canceled_registration -> ("Canceled_registration", [])
(* 17 *) | Canceled_theorem -> ("Canceled_theorem", [])
(* 18 *) | Case -> ("Case", [])
(* 19 *) | Case_list -> ("Case_list", [])
(* 20 *) | Choice_statement -> ("Choice_statement", [])
(* 21 *) | Cluster_registration -> ("Cluster_registration", [])
(* 22 *) | Collective_assumption -> ("Collective_assumption", [])
(* 23 *) | Compact_statement -> ("Compact_statement", [])
(* 24 *) | Conclusion -> ("Conclusion", [])
(* 25 *) | Conditional_definiens _-> ("Conditional_definiens", [])
(* 26 *) | Conditional_registration -> ("Conditional_registration", [])
(* 27 *) | Conditions -> ("Conditions", [])
(* 28 *) | Constant_definition -> ("Constant_definition", [])
(* 29 *) | Correctness_condition -> ("Correctness_condition", [])
(* 30 *) | Correctness_conditions -> ("Correctness_conditions", [])
(* 31 *) | Definiens _ -> ("Definiens", [])
(* 32 *) | Definition -> ("Definition", [])
(* 33 *) | Definition_item -> ("Definition_item", [])
(* 34 *) | Definition_number -> ("Definition_number", [])
(* 35 *) | Definitional_block -> ("Definitional_block", [])
(* 36 *) | Definitional_item -> ("Definitional_item", [])
(* 37 *) | Diffuse_conclusion -> ("Diffuse_conclusion" ,[])
(* 38 *) | Diffuse_statement -> ("Diffuse_statement", [])
(* 39 *) | Directive -> ("Directive", [])
(* 40 *) | Environment_declaration -> ("Environment_declaration", [])
(* 41 *) | Equating -> ("Equating", [])
(* 42 *) | Equating_list -> ("Equating_list", [])
(* 43 *) (* | Example -> ("Example", []) *)
(* 44 *) | Exemplification -> ("Exemplification", [])
(* 45 *) | Existential_assumption -> ("Existential_assumption", [])
(* 46 *) | Existential_registration -> ("Existential_registration", [])
(* 47 *) (* | Explicitly_qualified_variables -> ("Explicitly_qualified_variables", []) *)
(* 48 *) | Field_segment -> ("Field_segment", [])
(* 49 *) | Fields -> ("Fields", [])
(* 50 *) | File_name -> ("File_name", [])
(* 51 *) | Formula_expression -> ("Formula_expression", [])  
(* 52 *) | Functor_definition -> ("Functor_definition", [])
(* 53 *) | Functor_identifier -> ("Functor_identifier", [])
(* 54 *) | Functor_loci -> ("Functor_loci", [])
(* 55 *) | Functor_pattern -> ("Functor_pattern", [])
(* 56 *) | Functor_property -> ("Functor_property", [])
(* 57 *) (* | Functor_segment -> ("Functor_segment", []) *)
(* 58 *) | Functor_symbol -> ("Functor_symbol", [])
(* 59 *) | Functor_synonym -> ("Functor_synonym", [])
(* 60 *) | Functorial_registration -> ("Functorial_registration", [])
(* 61 *) | Generalization -> ("Generalization", [])
(* 62 *) (* | Identifier -> ("Identifier", []) *)
(* 63 *) (* | Implicitly_qualified_variables -> ("Implicitly_qualified_variables", []) *)
(* 64 *) | Iterative_equality -> ("Iterative_equality", [])
(* 65 *) | Justification -> ("Justification", [])
(* 66 *) | Label_identifier -> ("Label_identifier", [])
(* 67 *) | Left_functor_bracket -> ("Left_functor_bracket", [])
(* 68 *) | Library_directive lib -> ("Library_directive", [("kind", lib_to_attr lib)])
(* 69 *) | Library_reference -> ("Library_reference", [])
(* 70 *) | Library_scheme_reference -> ("Library_scheme_reference", [])
(* 71 *) | Linkable_statement -> ("Linkable_statement", [])
(* 72 *) | Local_reference -> ("Local_reference", [])
(* 73 *) | Local_scheme_reference -> ("Local_scheme_reference", [])
(* 74 *) | Loci -> ("Loci", [])
(* 75 *) | Loci_declaration -> ("Loci_declaration", [])
(* 76 *) | Locus -> ("Locus", [])
(* 77 *) | Mode_definition -> ("Mode_definition", [])
(* 78 *) | Mode_pattern -> ("Mode_pattern", [])
(* 79 *) | Mode_symbol -> ("Mode_symbol", [])
(* 80 *) | Mode_synonym -> ("Mode_synonym", [])
(* 81 *) | Notation_block -> ("Notation_block", [])
(* 82 *) | Notation_declaration -> ("Notation_declaration", []) 
(* 83 *) | Notation_item -> ("Notation_item", [])
(* 84 *) | Numeral -> ("Numeral", [])
(* 85 *) | Partial_definiens _ -> ("Partial_definiens", [])
(* 86 *) | Partial_definiens_list _ -> ("Partial_definiens_list", [])
(* 87 *) | Permissive_assumption -> ("Permissive_assumption", [])
(* 88 *) (* | Postqualification -> ("Postqualification", []) *)
(* 89 *) (* | Postqualified_variable -> ("Postqualified_variable", []) *)
(* 90 *) (* | Postqualifying_segment -> ("Postqualifying_segment", []) *)
(* 91 *) | Predicate_antonym -> ("Predicate_antonym", [])
(* 92 *) | Predicate_definition -> ("Predicate_definition", [])
(* 93 *) | Predicate_identifier -> ("Predicate_identifier", [])
(* 94 *) | Predicate_pattern -> ("Predicate_pattern", [])
(* 95 *) | Predicate_property -> ("Predicate_property", [])
(* 96 *) (* | Predicate_segment -> ("Predicate_segment", []) *)
(* 97 *) | Predicate_symbol -> ("Predicate_symbol", [])  
(* 98 *) | Predicate_synonym -> ("Predicate_synonym", [])
(* 99 *) | Private_definition -> ("Private_definition", [])
(* 100 *) (* | Private_definition_parameter -> ("Private_definition_parameter", []) *)
(* 101 *) | Private_functor_definition -> ("Private_functor_fefinition", [])
(* 102 *) | Private_functor_pattern -> ("Private_functor_pattern", [])
(* 103 *) | Private_predicate_definition -> ("Private_predicate_definition", [])
(* 104 *) | Private_predicate_pattern -> ("Private_predicate_pattern", [])
(* 105 *) | Proof -> ("Proof", [])
(* 106 *) | Proposition -> ("Proposition", [])
(* 107 *) (* | Qualification -> ("Qualification", []) *)
(* 108 *) (* | Qualified_segment -> ("Qualified_segment", []) *)
(* 109 *) | Qualified_variables -> ("Qualified_variables", [])
(* 110 *) (* | Quantified_formula_expression -> ("Quantified_formula_expression", []) *)
(* 111 *) (* | Radix_type -> ("Radix_type", []) *)
(* 112 *) | Reasoning -> ("Reasoning", [])
(* 113 *) | Reasoning_item -> ("Reasoning_item", [])
(* 114 *) | Redefinition_block -> ("Redefinition_block", [])
(* 115 *) | Reference -> ("Reference", [])
(* 116 *) | References -> ("References", [])
(* 117 *) | Registration_block -> ("Registration_block", [])
(* 118 *) | Registration_item -> ("Registration_item", [])
(* 119 *) | Requirement -> ("Requirement", [])
(* 120 *) | Requirement_directive -> ("Requirement_directive", [])
(* 121 *) | Reservation -> ("Reservation", [])
(* 122 *) (* | Reservation_segment -> ("Reservation_segment", []) *)
(* 123 *) (* | Reserved_identifiers -> ("Reserved_identifiers", []) *)
(* 124 *) | Right_functor_bracket -> ("Right_functor_bracket", [])
(* 125 *) | Scheme_block -> ("Scheme_block", [])
(* 126 *) | Scheme_conclusion -> ("Scheme_conclusion", [])
(* 127 *) | Scheme_identifier -> ("Scheme_identifier", [])
(* 128 *) | Scheme_item -> ("Scheme_item", [])  
(* 129 *) | Scheme_justification -> ("Scheme_justification", [])
(* 130 *) | Scheme_number -> ("Scheme_number", [])
(* 131 *) | Scheme_parameters -> ("Scheme_parameters", [])
(* 132 *) | Scheme_premise -> ("Scheme_premise", [])
(* 133 *) | Scheme_reference -> ("Scheme_reference", [])
(* 134 *) (* | Scheme_segemnt -> ("Scheme_segemnt", []) *)
(* 135 *) | Section -> ("Section", [])
(* 136 *) | Selector_symbol -> ("Selector_symbol", [])
(* 137 *) | Sentence -> ("Sentence", [])
(* 138 *) | Simple_definiens _ -> ("Simple_definiens", [])
(* 139 *) | Simple_justification -> ("Simple_justification", [])
(* 140 *) | Single_assumption -> ("Single_assumption", [])
(* 141 *) | Skeleton_item -> ("Skeleton_item", [])
(* 142 *) | Specification -> ("Specification", [])
(* 143 *) | Statement -> ("Statement", [])
(* 144 *) | Straightforward_justification -> ("Straightforward_justification", [])
(* 145 *) | Structure_definition -> ("Structure_definition", [])
(* 146 *) | Structure_symbol -> ("Structure_symbol", [])
(* 147 *) | Structure_type_expression -> ("Structure_type_expression", [])
(* 148 *) | Suppose -> ("Suppose", [])
(* 149 *) | Suppose_list -> ("Suppose_list", [])
(* 150 *) (* | Symbol -> ("Symbol", []) *)
(* 151 *) | Term_expression -> ("Term_expression", [])
(* 152 *) | Term_expression_list -> ("Term_expression_list", [])
(* 153 *) | Text_item -> ("Text_item", [])
(* 154 *) | Text_proper -> ("Text_proper", [])
(* 155 *) | Theorem -> ("Theorem", [])
(* 156 *) | Theorem_number -> ("Theorem_number", [])
(* 157 *) | Type_change_list -> ("Type_change_list", [])
(* 158 *) | Type_changing_statement -> ("Type_changing_statement", [])
(* 159 *) | Type_expression -> ("Type_expression", [])
(* 160 *) | Type_expression_list -> ("Type_expression_list", [])
(* 161 *) | Variable_identifer -> ("Variable_identifer", [])
(* 162 *) (* | Variables -> ("Variables", []) *)
(* 163 *) | Vocabulary_directive -> ("Vocabulary_directive", [])
(* 164 *) | Vocabulary_name -> ("Vocabulary_name", [])  

let open_ast_tag t =
  let (name, attr) = ast_miz_to_tag t in
  "<" ^ name ^ (attr_to_str attr) ^ ">"

let close_ast_tag t =
  let (name, _) = ast_miz_to_tag t in
  "</" ^ name ^ ">"

type nt_rec = {
  nt_type: ast_miz;
  nt_first_tk: t_token;
  nt_last_tk: t_token;
  mutable nt_sublist: nt_rec list;
  mutable nt_expanded: bool }
(* and *)
(*  t_ast = Nonterminal of nt_rec | Terminal of t_token *)

let create_rec t tk1 tk2 ls expanded = {
  nt_type = t;
  nt_first_tk = tk1;
  nt_last_tk = tk2;
  nt_sublist = ls;
  nt_expanded = expanded
}

(*
let rec print_ast indent t =
  let rec s_indent indent =
    if indent = 0 then "" else " " ^ (s_indent (indent - 1))
  in
  match t with
    Nonterminal nt -> print_string ((s_indent indent) ^ "Nonterminal\n")
  | Terminal tk -> Lexer.print_token tk
*)

let next_token tk =
  match tk.tk_next with
    Some tk1 -> tk1
  | None when tk.tk_type = Miz_tk_eof -> tk
  | _ -> raise Not_found

let rec first_non_ws (tk : Lexer.t_token) =
  match tk.tk_type with
  | Miz_tk_whitespace | Miz_tk_comment | Miz_tk_eol -> first_non_ws (next_token tk)
  |_ -> tk

let next_non_ws tk =
  first_non_ws (next_token tk)

let rec prev_non_ws tk =
  let rec f tk1 =
    match tk1.tk_type with
    | Miz_tk_whitespace | Miz_tk_comment | Miz_tk_eol -> f tk1.tk_prev
    |_ -> tk1
  in f tk.tk_prev

exception Syntax_error of int * int * string
let syntax_error line pos message = raise (Syntax_error (line,pos,message))
let syntax_error_tk tk message = syntax_error tk.tk_line tk.tk_pos message

(*
let rec tokens_to_txt tk_start tk_end =
  if tk_start == tk_end then "" else
  (if tk_start.tk_type = Miz_tk_eol then "\n" else tk_start.tk_text) ^ (tokens_to_txt (next_token tk_start) tk_end)
*)

let rec tokens_to_buf tk_start tk_end buf =
  if tk_start == tk_end then () else
    let () = if tk_start.tk_type = Miz_tk_eol then Buffer.add_string buf "\n" else Buffer.add_string buf tk_start.tk_text in
    tokens_to_buf (next_token tk_start) tk_end buf

let tokens_to_txt tk_start tk_end =
  let buf = Buffer.create 16 in
  let () = tokens_to_buf tk_start tk_end buf in
  Buffer.contents buf

let node_txt node =
  tokens_to_txt node.nt_first_tk node.nt_last_tk

let read_file_name tk =
  let () = if tk.tk_type <> Miz_tk_ident then syntax_error_tk tk "Expected file name" in
  let tk1 = next_non_ws tk in
  let node = create_rec File_name tk tk1 [] true in (node, tk1)

let read_ast_name kind tk =
  let (n1, tk1) = read_file_name tk in
  let node = create_rec kind tk tk1 [n1] true in
  (node, tk1)

(* Returns a list of nodes. *)
let rec read_list kind tk =
  let (n1, tk1) = read_ast_name kind tk in
  match tk1.tk_type with
  | Miz_semicolon -> ([n1], next_non_ws tk1)
  | Miz_comma -> let (u,v) = read_list kind (next_non_ws tk1) in (n1 :: u, v)
  | _ -> syntax_error_tk tk1 "Expected ',' or ';'"
  
let build_voc_directive tk =
  let (ls, tk1) = read_list Vocabulary_name (next_non_ws tk) in
  let node = create_rec Vocabulary_directive tk tk1 ls true in
  (node, tk1)

let build_req_directive tk =
  let (ls, tk1) = read_list Requirement (next_non_ws tk) in
  let node = create_rec Requirement_directive tk tk1 ls true in
  (node, tk1)

let build_lib_directive lib tk =
  let (ls, tk1) = read_list Article_name (next_non_ws tk) in
  let node = create_rec (Library_directive lib) tk tk1 ls true in
  (node, tk1)

let rec read_directives tk =
  let f (node, tk1) =
    (create_rec Directive tk tk1 [node] true, tk1) in
  let g n tk1 =
    let (u, v) = read_directives tk1 in (n :: u, v) in
  match tk.tk_type with
  | Miz_vocabularies  -> let (n, tk1) = f (build_voc_directive tk) in g n tk1
  | Miz_requirements  -> let (n, tk1) = f (build_req_directive tk) in g n tk1
  | Miz_notations     -> let (n, tk1) = f (build_lib_directive Notations tk) in g n tk1
  | Miz_constructors  -> let (n, tk1) = f (build_lib_directive Constructors tk) in g n tk1
  | Miz_registrations -> let (n, tk1) = f (build_lib_directive Registrations tk) in g n tk1
  | Miz_definitions   -> let (n, tk1) = f (build_lib_directive Definitions tk) in g n tk1
  | Miz_theorems      -> let (n, tk1) = f (build_lib_directive Theorems tk) in g n tk1
  | Miz_schemes       -> let (n, tk1) = f (build_lib_directive Schemes tk) in g n tk1
  | _ -> ([], tk)

let read_environ tk =
  let () = if tk.tk_type <> Miz_environ then (syntax_error tk.tk_line tk.tk_pos "expected environ") in
  let tk1 = next_non_ws tk in
  let (ls, tk2) = read_directives tk1 in
  let node = create_rec Environment_declaration tk tk2 ls true in
  (node, tk2)

(* Returns tk1 such that tk1.tk_type = t and level of of (case|suppose|hereby|now|proof|@proof)...end = 0. *)
(* definition|notation|registration ... end *)
let find_outer_ch tk t =
  let rec find_level level tk t =
    if level = 0 && tk.tk_type = t then tk else
    match tk.tk_type with
    | Miz_case -> find_level (level+1) (next_non_ws tk) t
    | Miz_suppose -> find_level (level+1) (next_non_ws tk) t
    | Miz_hereby -> find_level (level+1) (next_non_ws tk) t
    | Miz_now -> find_level (level+1) (next_non_ws tk) t
    | Miz_proof -> find_level (level+1) (next_non_ws tk) t
    | Miz_at_proof -> find_level (level+1) (next_non_ws tk) t
    | Miz_definition -> find_level (level+1) (next_non_ws tk) t
    | Miz_notation -> find_level (level+1) (next_non_ws tk) t
    | Miz_registration -> find_level (level+1) (next_non_ws tk) t
    | Miz_end -> find_level (level-1) (next_non_ws tk) t
    | Miz_tk_eof -> raise Not_found
    | _ -> find_level level (next_non_ws tk) t
  in
  find_level 0 tk t

let find_first_outer level tk =
  let rec find_level level tk =
    if level = 0 then tk else
    match tk.tk_type with
    | Miz_case -> find_level (level+1) (next_non_ws tk)
    | Miz_suppose -> find_level (level+1) (next_non_ws tk)
    | Miz_hereby -> find_level (level+1) (next_non_ws tk)
    | Miz_now -> find_level (level+1) (next_non_ws tk)
    | Miz_proof -> find_level (level+1) (next_non_ws tk)
    | Miz_at_proof -> find_level (level+1) (next_non_ws tk)
    | Miz_definition -> find_level (level+1) (next_non_ws tk)
    | Miz_notation -> find_level (level+1) (next_non_ws tk)
    | Miz_registration -> find_level (level+1) (next_non_ws tk)
    | Miz_end -> find_level (level-1) (next_non_ws tk)
    | Miz_tk_eof -> raise Not_found
    | _ -> find_level level (next_non_ws tk)
  in
  find_level level tk

let find_outer tk tk_end t =
  let rec find_level level tk t =
    if tk == tk_end then raise Not_found else
    if level = 0 && tk.tk_type = t then tk else
    match tk.tk_type with
    | Miz_case -> find_level (level+1) (next_non_ws tk) t
    | Miz_suppose -> find_level (level+1) (next_non_ws tk) t
    | Miz_hereby -> find_level (level+1) (next_non_ws tk) t
    | Miz_now -> find_level (level+1) (next_non_ws tk) t
    | Miz_proof -> find_level (level+1) (next_non_ws tk) t
    | Miz_at_proof -> find_level (level+1) (next_non_ws tk) t
    | Miz_definition -> find_level (level+1) (next_non_ws tk) t
    | Miz_notation -> find_level (level+1) (next_non_ws tk) t
    | Miz_registration -> find_level (level+1) (next_non_ws tk) t
    | Miz_end -> find_level (level-1) (next_non_ws tk) t
    | _ -> find_level level (next_non_ws tk) t
  in
  find_level 0 tk t

let find_outers tk tk_end ls =
  let rec find_level level tk ls =
    if tk == tk_end then raise Not_found else
    if level = 0 && (List.mem tk.tk_type ls) then tk else
    match tk.tk_type with
    | Miz_case -> find_level (level+1) (next_non_ws tk) ls
    | Miz_suppose -> find_level (level+1) (next_non_ws tk) ls
    | Miz_hereby -> find_level (level+1) (next_non_ws tk) ls
    | Miz_now -> find_level (level+1) (next_non_ws tk) ls
    | Miz_proof -> find_level (level+1) (next_non_ws tk) ls
    | Miz_at_proof -> find_level (level+1) (next_non_ws tk) ls
    | Miz_definition -> find_level (level+1) (next_non_ws tk) ls
    | Miz_notation -> find_level (level+1) (next_non_ws tk) ls
    | Miz_registration -> find_level (level+1) (next_non_ws tk) ls
    | Miz_end -> find_level (level-1) (next_non_ws tk) ls
    | _ -> find_level level (next_non_ws tk) ls
  in
  find_level 0 tk ls

(* Returns tk such that tk.tk_type = t and start_tk <= tk < end_tk. *)
let rec find_first start_tk t end_tk =
  if start_tk == end_tk then raise Not_found else
  if start_tk.tk_type = t then start_tk else
  find_first (next_token start_tk) t end_tk

let rec find_firsts start_tk ls end_tk =
  if start_tk == end_tk then raise Not_found else
  if List.mem start_tk.tk_type ls then start_tk else
  find_firsts (next_token start_tk) ls end_tk

let rec find_firsts_to_end start_tk ls  =
  if start_tk.tk_type = Miz_tk_eof then raise Not_found else
  if List.mem start_tk.tk_type ls then start_tk else
  find_firsts_to_end (next_token start_tk) ls

let find_outer_lc_rc tk_start tk_end t =
  let rec find tk level =
  if tk == tk_end then raise Not_found else
    match tk.tk_type with
    | Miz_lc -> find (next_non_ws tk) (level+1)
    | Miz_rc -> find (next_non_ws tk) (level-1)
    | tt when tt = t && level = 0 -> tk
    | _ -> find (next_non_ws tk) level
  in
  find tk_start 0

let find_matching_rp tk_start tk_end =
  let rec find tk level =
  if tk == tk_end then raise Not_found else
    match tk.tk_type with
    | Miz_lp -> find (next_non_ws tk) (level+1)
    | Miz_rp -> if level=1 then tk else find (next_non_ws tk) (level-1)
    | _ -> find (next_non_ws tk) level
  in
  find tk_start 0

let find_matching_rc tk_start =
  let rec find tk level =
  if tk.tk_type = Miz_tk_eof then raise Not_found else
    match tk.tk_type with
    | Miz_lc -> find (next_non_ws tk) (level+1)
    | Miz_rc -> if level=1 then tk else find (next_non_ws tk) (level-1)
    | _ -> find (next_non_ws tk) level
  in
  find tk_start 0

let read_text_item t tk =
  let tk1 = find_outer_ch tk Miz_semicolon in
  let tk2 = next_non_ws tk1 in
  let node = create_rec t tk tk2 [] false in
  (node, tk2)

let read_numeral tk =
  match tk.tk_type with
  | Miz_tk_numeral -> (create_rec Numeral tk (next_non_ws tk) [] true, next_non_ws tk)
  | _ -> syntax_error_tk tk "Expected numeral"

let l_read_numeral_opt (ls, tk) =
  match tk.tk_type with
  | Miz_tk_numeral -> (ls @ [create_rec Numeral tk (next_non_ws tk) [] true], next_non_ws tk)
  | _ -> ([], tk)

let is_selector_tk tk =
  match tk.tk_type with
  | Miz_tk_symbol q when q = Voc.Selector -> true
  | _ -> false

let is_structure_tk tk =
  match tk.tk_type with
  | Miz_tk_symbol q when q = Voc.Structure -> true
  | _ -> false

let read_selector_symbol tk =
 let () = if not (is_selector_tk tk) then syntax_error_tk tk "Expected selector symbol" in
  let node = create_rec Selector_symbol tk (next_non_ws tk) [] true in
  (node, next_non_ws tk)

let read_structure_symbol tk =
 let () = if not (is_structure_tk tk) then syntax_error_tk tk "Expected structure symbol" in
  let node = create_rec Structure_symbol tk (next_non_ws tk) [] true in
  (node, next_non_ws tk)

(** Variable-Identifier = Identifier *)
let read_variable_identifier tk =
  let tk1 = next_non_ws tk in
  create_rec Variable_identifer tk tk1 [] true

(** Scheme-Identifier = Identifier . *)
let read_scheme_identifer tk =
  match tk.tk_type with
  | Miz_tk_ident -> let tk_next = next_non_ws tk in (create_rec Scheme_identifier tk tk_next [] true, tk_next)
  | _ -> syntax_error_tk tk "Expected scheme identifier"

(** Label-Identifier = Identifier . *)
let read_label_identifer tk =
  match tk.tk_type with
  | Miz_tk_ident -> let tk_next = next_non_ws tk in (create_rec Label_identifier tk tk_next [] true, tk_next)
  | _ -> syntax_error_tk tk "Expected label identifier"

(** Local-Reference = Label-Identifier . *)
let read_local_reference tk =
  let (n, tk_next) = read_label_identifer tk in
  (create_rec Local_reference tk tk_next [n] true, tk_next)


(** Local-Scheme-Reference = Scheme-Identifier . *)
let read_local_scheme_reference tk =
  let (n, tk_next) = read_scheme_identifer tk in
  (create_rec Local_scheme_reference tk tk_next [n] true, tk_next)

(** Article-Name = File-Name . *)
let read_article_name tk =
  let (n, tk_next) = read_file_name tk in
  (create_rec Article_name tk tk_next [n] true, tk_next)

(** Scheme-Number = Numeral . *)
let read_scheme_number tk =
  let (n, tk_next) = read_numeral tk in
  (create_rec Scheme_number tk tk_next [n] true, tk_next)

(** Theorem-Number = Numeral . *)
let read_theorem_number tk =
  let (n, tk_next) = read_numeral tk in
  (create_rec Theorem_number tk tk_next [n] true, tk_next)

(** Definition-Number = Numeral . *)
let read_definition_number tk =
  let (n, tk_next) = read_numeral tk in
  (create_rec Definition_number tk tk_next [n] true, tk_next)

(** Library-Scheme-Reference = Article-Name ":" "sch" Scheme-Number . *)
let read_library_scheme_reference tk =
  let (n1, tk_colon) = read_article_name tk in
  let () = if tk_colon.tk_type <> Miz_colon then syntax_error_tk tk_colon "Expected ':'" in
  let tk_sch = next_non_ws tk_colon in
  let () = if tk_sch.tk_type <> Miz_sch then syntax_error_tk tk_sch "Expected sch" in
  let (n2, tk_next) = read_scheme_number (next_non_ws tk_sch) in
  (create_rec Library_scheme_reference tk tk_next [n1; n2] true, tk_next)

(** Scheme-Reference = Local-Scheme-Reference | Library-Scheme-Reference . *)
let read_scheme_reference tk =
  let tk1 = next_non_ws tk in
  if tk1.tk_type = Miz_colon then read_library_scheme_reference tk else read_local_scheme_reference tk

(** Library-Reference = Article-Name ":" ( Theorem-Number | "def" Definition-Number ) { "," ( Theorem-Number | "def" Definition-Number ) } . *)
let read_library_reference tk =
  let read_item tk =
    match tk.tk_type with
    | Miz_tk_numeral -> read_theorem_number tk
    | Miz_def -> read_definition_number (next_non_ws tk)
    | _ -> syntax_error_tk tk "Expected theorem number or def"
  in
  let rec read_list tk =
    let (n1, tk1) = read_item tk in
    let tk2 = next_non_ws tk1 in
    if tk1.tk_type = Miz_comma  && (tk2.tk_type = Miz_tk_numeral || tk2.tk_type = Miz_def) then let (u, v) = read_list tk2 in (n1 :: u, v) else ([n1], tk1)
  in
  let (n, tk_colon) = read_article_name tk in
  let () = if tk_colon.tk_type <> Miz_colon then syntax_error_tk tk_colon "Expected ':'" in
  let (ls, tk_next) = read_list (next_non_ws tk_colon) in
  (create_rec Library_reference tk tk_next (n :: ls) true, tk_next)

(** Reference = Local-Reference | Library-Reference . *)
let read_reference tk =
  let tk1 = next_non_ws tk in
  if tk1.tk_type = Miz_colon then read_library_reference tk else read_local_reference tk

(** References = Reference { "," Reference } . *)
let read_references tk =
  let rec read_list tk =
    let (n1, tk1) = read_reference tk in
    if tk1.tk_type = Miz_comma then let (u, v) = read_list (next_non_ws tk1) in (n1 :: u, v) else ([n1], tk1)
  in
  let (ls, tk_next) = read_list tk in
  (create_rec References tk tk_next ls true, tk_next)

(** Scheme-Justification = "from" Scheme-Reference [ "(" References ")" ] . *)
let read_scheme_justification tk =
  let () = if tk.tk_type <> Miz_from then syntax_error_tk tk "Expected from" in
  let (n1, tk1) = read_scheme_reference (next_non_ws tk) in
  if tk1.tk_type = Miz_lp then
    let (n2, tk2) = read_references (next_non_ws tk1) in
    let () = if tk2.tk_type <> Miz_rp then syntax_error_tk tk2 "Expected ')'" in
    (create_rec Scheme_justification tk (next_non_ws tk2) [n1; n2] true, next_non_ws tk2)
  else
    (create_rec Scheme_justification tk tk1 [n1] true, tk1)

(** Straightforward-Justification = [ "by" References ] . *)
let read_straightforward_justification tk =
  if tk.tk_type = Miz_by then
    let (n, tk_next) = read_references (next_non_ws tk) in
    (create_rec Straightforward_justification tk tk_next [n] true, tk_next)
  else
    (create_rec Straightforward_justification tk tk [] true, tk)

(** Simple-Justification = Straightforward-Justification | Scheme-Justification . *)
let read_simple_justification tk =
  match tk.tk_type with
  | Miz_from -> read_scheme_justification tk
  | _ -> read_straightforward_justification tk

(** Locus = Variable-Identifier *)
let read_locus tk =
  let () = if tk.tk_type <> Miz_tk_ident then syntax_error_tk tk "Expected identifier" in
  let n = read_variable_identifier tk in
  let node = create_rec Locus tk n.nt_last_tk [n] true in
  (node, n.nt_last_tk)

let rec read_locus_list tk =
  let (n, tk1) = read_locus tk in
  match tk1.tk_type with
  | Miz_comma -> let (u, v) = read_locus_list (next_non_ws tk1) in (n :: u, v)
  | _ -> ([n], tk1)

(** Loci = Locus { "," Locus } . *)
let read_loci (ls, tk) =
  let (ls1, tk1) = read_locus_list tk in
  let node = create_rec Loci tk tk1 ls1 true in
  (ls @ [node], tk1)

let l_read_loci_opt (ls, tk) =
  match tk.tk_type with
  | Miz_tk_ident -> read_loci (ls, tk)
  | _ -> (ls, tk)

(** Functor-Loci = Locus | "(" Loci ")" . *)
let read_functor_loci tk =
  match tk.tk_type with
  | Miz_tk_ident -> let (n, tk1) = read_locus tk in (create_rec Functor_loci tk tk1 [n] true, tk1)
  | Miz_lp -> let (ls, tk1) = read_loci ([],next_non_ws tk) in 
              let () = if tk1.tk_type <> Miz_rp then syntax_error_tk tk1 "Expected ')'" in
              (create_rec Functor_loci tk (next_non_ws tk1) ls true, next_non_ws tk1)
  | _ -> syntax_error_tk tk "Expected locus or '('"

let l_read_functor_loci_opt (ls, tk) =
  match tk.tk_type with
  | Miz_tk_ident -> let (n, tk1) = read_locus tk in (ls @ [create_rec Functor_loci tk tk1 [n] true], tk1)
  | Miz_lp -> let (ls1, tk1) = read_loci ([],next_non_ws tk) in 
              let () = if tk1.tk_type <> Miz_rp then syntax_error_tk tk1 "Expected ')'" in
              (ls @ [create_rec Functor_loci tk (next_non_ws tk1) ls1 true], next_non_ws tk1)
  | _ -> (ls, tk)

(** Mode-Symbol = Symbol | "set" . *)
let read_mode_symbol tk =
  let () =
    match tk.tk_type with
    | Miz_tk_symbol q when q = Voc.Mode -> ()
    | Miz_set -> ()
    | _ -> syntax_error_tk tk ("Expected mode symbol:" ^ tk.tk_text)
  in
  let tk1 = next_non_ws tk in
  let node = create_rec Mode_symbol tk tk1 [] true in
  (node, tk1)

(** Predicate-Symbol = Symbol . *)
let read_predicate_symbol tk =
  let () =
    match tk.tk_type with
    | Miz_tk_symbol q when q = Voc.Predicate -> ()
    | Miz_eq -> ()
    | _ -> syntax_error_tk tk ("Expected predicate symbol:" ^ tk.tk_text)
  in
  let tk1 = next_non_ws tk in
  let node = create_rec Predicate_symbol tk tk1 [] true in
  (node, tk1)

(** Attribute-Symbol = Symbol . *)
let read_attribute_symbol tk =
  let () =
    match tk.tk_type with
    | Miz_tk_symbol q when q = Voc.Attribute -> ()
    | _ -> syntax_error_tk tk ("Expected attribute symbol:" ^ tk.tk_text)
  in
  let tk1 = next_non_ws tk in
  let node = create_rec Attribute_symbol tk tk1 [] true in
  (node, tk1)

(** Functor-Symbol = Symbol . *)
let read_functor_symbol tk =
  let () =
    match tk.tk_type with
    | Miz_tk_symbol q when (match q  with Voc.Functor _ -> true | _ -> false) -> ()
    | _ -> syntax_error_tk tk ("Expected functor symbol:" ^ tk.tk_text)
  in
  let tk1 = next_non_ws tk in
  let node = create_rec Functor_symbol tk tk1 [] true in
  (node, tk1)

let l_read_functor_symbol (ls, tk) = let (n, tk_next) = read_functor_symbol tk in (ls @ [n], tk_next)
let l_read_predicate_symbol (ls, tk) = let (n, tk_next) = read_predicate_symbol tk in (ls @ [n], tk_next)
let l_read_attribute_symbol (ls, tk) = let (n, tk_next) = read_attribute_symbol tk in (ls @ [n], tk_next)
let l_read_numeral (ls, tk) = let (n, tk_next) = read_numeral tk in (ls @ [n], tk_next)

(** Reservation = "reserve" Reservation-Segment { "," Reservation-Segment } ";" . *) 
let read_reservation tk =
  let tk1 = find_outer_ch tk Miz_semicolon in
  let tk2 = next_non_ws tk1 in
  let node = create_rec Reservation tk tk2 [] false in
  (node, tk2)

let l_read_kw_non_opt (ls, tk) =
  match tk.tk_type with
  | Miz_non -> (ls, next_non_ws tk)
  | _ -> (ls, tk)

let read_adjective tk =
  match tk.tk_type with
  | Miz_non -> let (n, tk_next) = read_attribute_symbol (next_non_ws tk) in
               (create_rec Adjective tk tk_next [n] true, tk_next)
  | _ -> let (n, tk_next) = read_attribute_symbol tk in
               (create_rec Adjective tk tk_next [n] true, tk_next)

(** Adjective = [ "non" ] Attribute-Symbol . *)
let l_read_adjective (ls, tk) =
  let (ls1, tk_next) =l_read_attribute_symbol (l_read_kw_non_opt (ls, tk)) in
  (ls @ [create_rec Adjective tk tk_next ls1 true], tk_next)

let rec read_adjective_list tk =
  match tk.tk_type with
  | Miz_non -> let (n, tk1) = read_adjective tk in let (u, v) = read_adjective_list tk1 in (n :: u, v)
  | Miz_tk_symbol q when q = Voc.Attribute -> let (n, tk1) = read_adjective tk in let (u, v) = read_adjective_list tk1 in (n :: u, v)
  | _ -> ([], tk)

(** Adjective-Cluster = { Adjective } . *)
let read_adjective_cluster tk =
  let (ls1, tk1) = read_adjective_list tk in
  (create_rec Adjective_cluster tk tk1 ls1 true, tk1)

(** Adjective-Cluster = { Adjective } . *)
let l_read_adjective_cluster (ls, tk) =
  let (ls1, tk1) = read_adjective_list tk in
  (ls @ [create_rec Adjective_cluster tk tk1 ls1 true], tk1)

(* Structure-Type-Expression = "(" Structure-Type-Expression ")" | Adjective-Cluster Structure-Symbol [ "over" Term-Expression-List ] . *)

let rec skip_lp_seq tk =
  match tk.tk_type with
  | Miz_lp -> skip_lp_seq (next_non_ws tk)
  | _ -> tk

(* TODO: Very slow *)
(** Ancestors = Structure-Type-Expression { "," Structure-Type-Expression } . *)
let expand_ancestors node =
  let is_next tk =
    if tk.tk_type = Miz_comma then
      let tk1 = skip_lp_seq (next_non_ws tk) in
      let (_, tk1) = read_adjective_cluster tk1 in
      if is_structure_tk tk1 then (next_non_ws tk1).tk_type <> Miz_lpound else false
    else false
  in
  let rec find_next tk =
    if tk == node.nt_last_tk then tk
    else if is_next tk then tk
    else find_next (next_non_ws tk)
  in
  let read_item tk =
    if tk.tk_type = Miz_lp then
      let tk1 = find_matching_rp tk node.nt_last_tk in
      let n = create_rec Structure_type_expression tk (next_non_ws tk1) [] false in
      (n, next_non_ws tk1)
    else
     let (n1, tk1) = read_adjective_cluster tk in
     let (n2, tk2) = read_structure_symbol tk1 in
     if tk2.tk_type = Miz_over then
       let tk_next = find_next tk2 in
       let n3 = create_rec Term_expression_list (next_non_ws tk2) tk_next [] false in
       (create_rec Structure_type_expression tk tk_next [n1; n2; n3] true, tk_next)
     else
     (create_rec Structure_type_expression tk tk2 [n1; n2] true, tk2)
  in
  let rec read_list tk =
    let (n, tk1) = read_item tk in
    if tk1 == node.nt_last_tk then [n]
    else let ls = read_list (next_non_ws tk1) in n :: ls
  in
  let tk = node.nt_first_tk in
  let ls = read_list tk in
  (node.nt_sublist <- ls; node.nt_expanded <- true)

(** Loci-Declaration = "let" Qualified-Variables [ "such" Conditions ] ";" . *)
let read_loci_declaration tk =
  let tk_semi = try find_outer_ch tk Miz_semicolon with Not_found -> syntax_error_tk tk "There is no ending ';'" in
  let tk_next = next_non_ws tk_semi in
  (create_rec Loci_declaration tk tk_next [] false, tk_next)

let is_adjective_cluster tk_start tk_end =
  let (_, tk_next) = read_adjective_list tk_start in
  tk_next == tk_end

let l_read_kw_end (ls, tk) = if tk.tk_type <> Miz_end then syntax_error_tk tk "Expected end" else (ls, next_non_ws tk)
let l_read_kw_registration (ls, tk) = if tk.tk_type <> Miz_registration then syntax_error_tk tk "Expected registration" else (ls, next_non_ws tk)
let l_read_kw_notation (ls, tk) = if tk.tk_type <> Miz_notation then syntax_error_tk tk "Expected notation" else (ls, next_non_ws tk)

let l_read_loci_declaration (ls, tk) =
  let (n, tk_next) = read_loci_declaration tk in
  (ls @ [n], tk_next)

(** Structure-Definition = "struct" [ "(" Ancestors ")" ] Structure-Symbol [ "over" Loci ] "(#" Fields "#)" ";" . *)
let read_structure_definition tk =
  let tk1 = find_outer_ch tk Miz_semicolon in
  let tk2 = next_non_ws tk1 in
  let node = create_rec Structure_definition tk tk2 [] false in
  (node, tk2)

(** Attribute-Pattern = Locus "is" Attribute-Symbol . *)
let l_read_attribute_pattern (ls, tk) =
  let (n1, tk1) = read_locus tk in
  let () = if tk1.tk_type <> Miz_is then syntax_error_tk tk1 "Expected is" in
  let (n2, tk_next) = read_attribute_symbol (next_non_ws tk1) in
  (ls @ [create_rec Attribute_pattern tk tk_next [n1; n2] true], tk_next)

(** Predicate-Pattern = [ Loci ] Predicate-Symbol [ Loci ] . *)
let l_read_predicate_pattern (ls, tk) = let (ls1, tk1) = l_read_loci_opt (l_read_predicate_symbol (l_read_loci_opt (ls, tk))) in
  (ls @ [create_rec Predicate_pattern tk tk1 ls1 true], tk1)

(** Left-Functor-Bracket = Symbol . *)
let read_left_functor_bracket tk =
  let () =
    match tk.tk_type with
    | Miz_tk_symbol q when q = Voc.Left_functor_bracket -> ()
    | Miz_lc -> ()
    | Miz_lb -> ()
    | _ -> syntax_error_tk tk ("Expected left functor bracket:" ^ tk.tk_text)
  in
  let tk1 = next_non_ws tk in
  let node = create_rec Left_functor_bracket tk tk1 [] true in
  (node, tk1)

let l_read_left_functor_bracket (ls, tk) = let (n, tk_next) = read_left_functor_bracket tk in (ls @ [n], tk_next)

(** Right-Functor-Bracket = Symbol . *)
let read_right_functor_bracket tk =
  let () =
    match tk.tk_type with
    | Miz_tk_symbol q when q = Voc.Right_functor_bracket -> ()
    | Miz_rc -> ()
    | Miz_rb -> ()
    | _ -> syntax_error_tk tk ("Expected right functor bracket:" ^ tk.tk_text)
  in
  let tk1 = next_non_ws tk in
  let node = create_rec Right_functor_bracket tk tk1 [] true in
  (node, tk1)

let l_read_right_functor_bracket (ls, tk) = let (n, tk_next) = read_right_functor_bracket tk in (ls @ [n], tk_next)

(** Functor-Pattern = [ Functor-Loci ] Functor-Symbol [ Functor-Loci ] | Left-Functor-Bracket Loci Right-Functor-Bracket . *)
let l_read_functor_pattern (ls, tk) =
  let (ls1, tk1) =
    match tk.tk_type with
    | Miz_lc | Miz_lb -> l_read_right_functor_bracket (read_loci (l_read_left_functor_bracket (ls, tk)))
    | Miz_tk_symbol q when q = Voc.Left_functor_bracket -> 
                    l_read_right_functor_bracket (read_loci (l_read_left_functor_bracket (ls, tk)))
    | _ ->
        (* l_read_functor_loci_opt (l_read_functor_symbol (l_read_functor_loci_opt (ls, tk))) *)
        let (u, v) = l_read_functor_symbol (l_read_functor_loci_opt (ls, tk)) in
        (* let () = print_endline (string_of_int (List.length u)) in *)
        (* let () = print_token v in *)
        (* let () = print_token (List.hd u).nt_first_tk in *)
        (*let () = print_token (List.nth u 1).nt_first_tk in
        let () = print_token (List.nth u 1).nt_last_tk in*)
        let (u, v) = l_read_functor_loci_opt (u, v) in
        (*let () = print_endline "----" in
        let () = print_endline (string_of_int (List.length u)) in
        let () = print_token v in *)
        (u, v)
  in (ls @ [create_rec Functor_pattern tk tk1 ls1 true], tk1)

(** Mode-Pattern = Mode-Symbol [ "of" Loci ] . *)
let read_mode_pattern tk =
  let (n, tk1) = read_mode_symbol tk in
  if tk1.tk_type = Miz_of then
    let (loci_list, tk2) = read_loci ([], (next_non_ws tk1)) in
    let node = create_rec Mode_pattern tk tk2 (n :: loci_list) true in
    (node, tk2)
  else
    let node = create_rec Mode_pattern tk tk1 [n] true in
    (node, tk1)

let rec skip_correctness_condition_list tk =
  match tk.tk_type with
  | Miz_existence
  | Miz_uniqueness
  | Miz_coherence
  | Miz_compatibility
  | Miz_consistency -> let tk1 = next_non_ws (find_outer_ch tk Miz_semicolon) in skip_correctness_condition_list tk1
  | _ -> tk

let skip_correctness_conditions tk =
  let tk1 = skip_correctness_condition_list tk in
  match tk1.tk_type with
  | Miz_correctness -> next_non_ws (find_outer_ch tk1 Miz_semicolon)
  | _ -> tk1

(** Correctness-Condition = ( "existence" | "uniqueness" | "coherence" | "compatibility" | "consistency" ) Justification ";" . *)
let rec read_correctness_condition_list tk =
  let f () =
    let tk_semi = try find_outer_ch tk Miz_semicolon with Not_found -> syntax_error_tk tk "There is no ending ';'" in
    let n_j = create_rec Justification (next_non_ws tk) tk_semi [] false in
    let tk_next = next_non_ws tk_semi in
    let n = create_rec Correctness_condition tk tk_next [n_j] true in
    let (u, v) = read_correctness_condition_list tk_next in
    (n :: u, v)
  in
  match tk.tk_type with
  | Miz_existence
  | Miz_uniqueness
  | Miz_coherence
  | Miz_compatibility
  | Miz_consistency -> f ()
  | _ -> ([], tk)

(** Correctness-Conditions = { Correctness-Condition } [ "correctness" Justification ";" ] . *)
let read_correctness_conditions tk =
  let (ls, tk_next) = read_correctness_condition_list tk in
  match tk_next.tk_type with
  | Miz_correctness -> let tk_semi = find_outer_ch tk_next Miz_semicolon in
                       let n = create_rec Justification (next_non_ws tk_next) tk_semi [] false in
                       (create_rec Correctness_conditions tk (next_non_ws tk_semi) (ls @ [n]) true, next_non_ws tk_semi)
  | _ -> (create_rec Correctness_conditions tk tk_next ls true, tk_next)

(** Canceled-Definition = "canceled" [ Numeral ] ";" . *)
let read_canceled_definition tk =
  let () = if tk.tk_type <> Miz_canceled then syntax_error_tk tk "Expected canceled" in
  let (ls, tk_semi) = l_read_numeral_opt ([], next_non_ws tk) in
  let () = if tk_semi.tk_type <> Miz_semicolon then syntax_error_tk tk_semi "Expected ';'" in
  let tk_next = next_non_ws tk_semi in
  let node = create_rec Canceled_definition tk tk_next ls true in
  (node, tk_next)

(** Canceled-Registration = "canceled" [ Numeral ] ";" . *)
let read_canceled_registration tk =
  let () = if tk.tk_type <> Miz_canceled then syntax_error_tk tk "Expected canceled" in
  let (ls, tk_semi) = l_read_numeral_opt ([], next_non_ws tk) in
  let () = if tk_semi.tk_type <> Miz_semicolon then syntax_error_tk tk_semi "Expected ';'" in
  let tk_next = next_non_ws tk_semi in
  let node = create_rec Canceled_registration tk tk_next ls true in
  (node, tk_next)

(** Canceled-Theorem = "canceled" [ Numeral ] ";" . *)
let read_canceled_theorem tk =
  let () = if tk.tk_type <> Miz_canceled then syntax_error_tk tk "Expected canceled" in
  let (ls, tk_semi) = l_read_numeral_opt ([], next_non_ws tk) in
  let () = if tk_semi.tk_type <> Miz_semicolon then syntax_error_tk tk_semi "Expected ';'" in
  let tk_next = next_non_ws tk_semi in
  let node = create_rec Canceled_theorem tk tk_next ls true in
  (node, tk_next)

(** Existential-Registration = "cluster" Adjective-Cluster Type-Expression ";" Correctness-Conditions . *)
let read_existential_registration tk tk_semi =
  let (n1, tk1) = read_adjective_cluster (next_non_ws tk) in
  let n2 = create_rec Type_expression tk1 tk_semi [] false in
  let (n3, tk_next) = read_correctness_conditions (next_non_ws tk_semi) in
  let node = create_rec Existential_registration tk tk_next [n1; n2; n3] true in
  (node, tk_next)

(** Conditional-Registration = "cluster" Adjective-Cluster "->" Adjective-Cluster Type-Expression ";" Correctness-Conditions . *)
let read_conditional_registration tk tk_semi =
  let (n1, tk1) = read_adjective_cluster (next_non_ws tk) in
  let () = if tk1.tk_type <> Miz_imp then syntax_error_tk tk1 "Expected '->'" in
  let (n2, tk2) = read_adjective_cluster (next_non_ws tk1) in
  let n3 = create_rec Type_expression tk2 tk_semi [] false in
  let (n4, tk_next) = read_correctness_conditions (next_non_ws tk_semi) in
  let node = create_rec Conditional_registration tk tk_next [n1; n2; n3; n4] true in
  (node, tk_next)

let read_type_expression_opt ls tk tk_end = if tk == tk_end then ls else ls @ [create_rec Type_expression tk tk_end [] false]

(** Functorial-Registration = "cluster" Term-Expression "->" Adjective-Cluster [ Type-Expression ]  ";" Correctness-Conditions . *)
let read_functorial_registration tk tk_imp tk_semi =  
  let n1 = create_rec Term_expression (next_non_ws tk) tk_imp [] false in
  let (n2, tk1) = read_adjective_cluster (next_non_ws tk_imp) in
  let ls = read_type_expression_opt [n1; n2] tk1 tk_semi in
  let (n3, tk_next) = read_correctness_conditions (next_non_ws tk_semi) in
  let node = create_rec Functorial_registration tk tk_next (ls @ [n3]) true in
  (node, tk_next)

(** Cluster-Registration = Existential-Registration | Conditional-Registration | Functorial-Registration | Canceled-Registration . *)
let read_cluster_registration tk =
  let f (n, tk_next) = (create_rec Cluster_registration tk tk_next [n] true, tk_next) in
  if tk.tk_type = Miz_canceled then f (read_canceled_registration tk)
  else
    let tk_semi = try find_outer_ch tk Miz_semicolon with Not_found -> syntax_error_tk tk "There is no ending ';'" in
    let () = if tk.tk_type <> Miz_cluster then syntax_error_tk tk "Expected cluster" in
    try
      let tk_imp = find_first tk Miz_imp tk_semi in
      if is_adjective_cluster (next_non_ws tk) tk_imp then f (read_conditional_registration tk tk_semi)  else f (read_functorial_registration tk tk_imp tk_semi)
    with Not_found -> f (read_existential_registration tk tk_semi)

let l_read_cluster_registration (ls, tk) = let (n, tk_next) = read_cluster_registration tk in (ls @ [n], tk_next)

let rec l_read_decl_reg_list (ls, tk) =
  match tk.tk_type with
  | Miz_let -> l_read_decl_reg_list (l_read_loci_declaration (ls, tk))
  | Miz_cluster | Miz_canceled -> l_read_decl_reg_list (l_read_cluster_registration (ls, tk))
  | _ -> (ls, tk)

(** Registration-Block = "registration" { ( Loci-Declaration | Cluster-Registration ) } "end" . *)
let read_registration_block tk =
  let (ls, tk_next) = l_read_kw_end (l_read_decl_reg_list (l_read_kw_registration ([], tk))) in
  (create_rec Registration_block tk tk_next ls true, tk_next)

(** Notation-Declaration = Attribute-Synonym | Attribute-Antonym | Functor-Synonym | Mode-Synonym | Predicate-Synonym | Predicate-Antonym . *)
let read_notation_declaration tk =
  let tk_semi = try find_outer_ch tk Miz_semicolon with Not_found -> syntax_error_tk tk "There is no ending ';'" in
  let node = create_rec Notation_declaration tk (next_non_ws tk_semi) [] false in
  (node, next_non_ws tk_semi)

let l_read_notation_declaration (ls, tk) = let (n, tk_next) = read_notation_declaration tk in (ls @ [n], tk_next)

let rec l_read_decl_not_list (ls, tk) =
  match tk.tk_type with
  | Miz_let -> l_read_decl_not_list (l_read_loci_declaration (ls, tk))
  | Miz_antonym | Miz_synonym -> l_read_decl_not_list (l_read_notation_declaration (ls, tk))
  | _ -> (ls, tk)

(** Notation-Block = "notation" { ( Loci-Declaration | Notation-Declaration ) } "end" . *)
let read_notation_block tk =
  let (ls, tk_next) = l_read_kw_end (l_read_decl_not_list (l_read_kw_notation ([], tk))) in
  (create_rec Notation_block tk tk_next ls true, tk_next)

(** Functor-Property = ( "commutativity" | "idempotence" | "involutiveness" | "projectivity" ) Justification ";" . *)
let rec read_functor_property_list tk =
  let f () =
    let tk_semi = try find_outer_ch tk Miz_semicolon with Not_found -> syntax_error_tk tk "There is no ending ';'" in
    let n_j = create_rec Justification (next_non_ws tk) tk_semi [] false in
    let tk_next = next_non_ws tk_semi in
    let n = create_rec Functor_property tk tk_next [n_j] true in
    let (u, v) = read_functor_property_list tk_next in
    (n :: u, v)
  in
  match tk.tk_type with
  | Miz_commutativity
  | Miz_idempotence
  | Miz_involutiveness
  | Miz_projectivity -> f ()
  | _ -> ([], tk)

(** Predicate-Property = ( "symmetry" | "asymmetry" | "connectedness" | "reflexivity" | "irreflexivity" ) Justification ";" . *)
let rec read_predicate_property_list tk =
  let f () =
    let tk_semi = try find_outer_ch tk Miz_semicolon with Not_found -> syntax_error_tk tk "There is no ending ';'" in
    let n_j = create_rec Justification (next_non_ws tk) tk_semi [] false in
    let tk_next = next_non_ws tk_semi in
    let n = create_rec Predicate_property tk tk_next [n_j] true in
    let (u, v) = read_predicate_property_list tk_next in
    (n :: u, v)
  in
  match tk.tk_type with
  | Miz_symmetry
  | Miz_asymmetry
  | Miz_connectedness
  | Miz_reflexivity
  | Miz_irreflexivity -> f ()
  | _ -> ([], tk)

(** Mode-Definition = "mode" Mode-Pattern ( [ Specification ] [ "means" Definiens ] ";" Correctness-Conditions | "is" Type-Expression ";" ) . *)
let read_mode_definition tk =
  let read_rest ls tk =
    let tk_semi = try find_outer_ch tk Miz_semicolon with Not_found -> syntax_error_tk tk "There is no ending ';'" in
    let read_rest_is ls tk =(ls @ [create_rec Type_expression tk tk_semi [] false], next_non_ws tk_semi) in
    let read_rest_correctness ls tk = let (n, tk1) = read_correctness_conditions tk in (ls @ [n], tk1) in
    let read_rest_means ls tk = read_rest_correctness (ls @ [create_rec (Definiens Sentence) tk tk_semi [] false]) (next_non_ws tk_semi) in
    let read_rest_imp ls tk =
      try
        let tk_means = find_outer tk tk_semi Miz_means in
        read_rest_means (ls @ [create_rec Specification tk tk_means [] false]) tk_means
      with Not_found -> read_rest_correctness (ls @ [create_rec Specification tk tk_semi [] false]) tk_semi
    in
    match tk.tk_type with
    | Miz_is -> read_rest_is ls (next_non_ws tk)
    | Miz_means -> read_rest_means ls (next_non_ws tk)
    | Miz_imp -> read_rest_imp ls tk
    | Miz_semicolon -> read_rest_correctness ls (next_non_ws tk)
    | _ -> syntax_error_tk tk "Expected '->' or 'is'"
  in
  let (n1, tk1) = read_mode_pattern (next_non_ws tk) in
  let (ls, tk_next) = read_rest [n1] tk1 in
  (create_rec Mode_definition tk tk_next ls true, tk_next)

(** Specification = "->" Type-Expression . *)
let expand_specification node =
  let tk = node.nt_first_tk in
  let () = if tk.tk_type <> Miz_imp then syntax_error_tk tk "Expected '->'" in
  let n = create_rec Type_expression (next_non_ws tk) node.nt_last_tk [] false in
  (node.nt_sublist <- [n]; node.nt_expanded <- true)

(** Functor-Definition = "func" Functor-Pattern [ Specification ] [ ( "means" | "equals" ) Definiens ] ";" Correctness-Conditions { Functor-Property } . *)
let read_functor_definition tk =
  let () = if tk.tk_type <> Miz_func then syntax_error_tk tk "Expected func" in
  let tk_semi = try find_outer_ch tk Miz_semicolon with Not_found -> syntax_error_tk tk "There is no ending ';'" in
  let read_fp_list (ls, tk) = let (ls1, tk1) = read_functor_property_list tk in (ls @ ls1, tk1) in
  let read_correctness (ls,tk) = let (n, tk1) = read_correctness_conditions tk in (ls @ [n], tk1) in
  let read_def (ls, tk) =
    match tk.tk_type with
    | Miz_means | Miz_equals -> (ls @ [create_rec (Definiens Term_expression) tk tk_semi [] false], next_non_ws tk_semi)
    | Miz_semicolon -> (ls, next_non_ws tk_semi)
    | _ -> syntax_error_tk tk "Expected means, equals or ';'"
  in
  let read_spec (ls, tk) =
    if tk.tk_type = Miz_imp then
      let tk_next = try find_firsts tk [Miz_means; Miz_equals] tk_semi with Not_found -> tk_semi in
      let n = create_rec Specification tk tk_next [] false in (ls @ [n], tk_next)
    else (ls, tk)
  in
  let (ls, tk_next) = read_fp_list (read_correctness (read_def (read_spec (l_read_functor_pattern ([], next_non_ws tk))))) in
  let node = create_rec Functor_definition tk tk_next ls true in
  (node, tk_next)

(** Predicate-Definition = "pred" Predicate-Pattern [ "means" Definiens ] ";" Correctness-Conditions  { Predicate-Property } . *)
let read_predicate_definition tk =
  let () = if tk.tk_type <> Miz_pred then syntax_error_tk tk "Expected pred" in
  let tk_semi = try find_outer_ch tk Miz_semicolon with Not_found -> syntax_error_tk tk "There is no ending ';'" in
  let read_pp_list (ls, tk) = let (ls1, tk1) = read_predicate_property_list tk in (ls @ ls1, tk1) in
  let read_correctness (ls,tk) = let (n, tk1) = read_correctness_conditions tk in (ls @ [n], tk1) in
  let read_def (ls, tk) =
    match tk.tk_type with
    | Miz_means -> (ls @ [create_rec (Definiens Sentence) tk tk_semi [] false], next_non_ws tk_semi)
    | Miz_semicolon -> (ls, next_non_ws tk_semi)
    | _ -> syntax_error_tk tk "Expected means or ';'"
  in
  let (ls, tk_next) = read_pp_list (read_correctness (read_def (l_read_predicate_pattern ([], next_non_ws tk)))) in
  let node = create_rec Predicate_definition tk tk_next ls true in
  (node, tk_next)

(** Attribute-Definition = "attr" Attribute-Pattern "means" Definiens ";" Correctness-Conditions . *)
let read_attribute_definition tk =
  let () = if tk.tk_type <> Miz_attr then syntax_error_tk tk "Expected attr" in
  let tk_semi = try find_outer_ch tk Miz_semicolon with Not_found -> syntax_error_tk tk "There is no ending ';'" in
  let read_correctness (ls,tk) = let (n, tk1) = read_correctness_conditions tk in (ls @ [n], tk1) in
  let read_def (ls, tk) =
    match tk.tk_type with
    | Miz_means -> (ls @ [create_rec (Definiens Sentence) tk tk_semi [] false], next_non_ws tk_semi)
    | Miz_semicolon -> (ls, next_non_ws tk_semi)
    | _ -> syntax_error_tk tk "Expected means or ';'"
  in
  let (ls, tk_next) = read_correctness (read_def (l_read_attribute_pattern ([], next_non_ws tk))) in
  let node = create_rec Attribute_definition tk tk_next ls true in
  (node, tk_next)

let read_auxiliary_item tk =
  let tk1 = find_outer_ch tk Miz_semicolon in
  let tk2 = next_non_ws tk1 in
  let node = create_rec Auxiliary_item tk tk2 [] false in
  (node, tk2)

(** Conditions = "that" Proposition { "and" Proposition } . *)
let expand_conditions start_tk end_tk =
  let () = if start_tk.tk_type <> Miz_that then syntax_error_tk start_tk "Expected that" in
  let tk = next_non_ws start_tk in
  let rec read_list tk =
    try
      let tk_and = find_first tk Miz_and end_tk in
      let node_p = create_rec Proposition tk tk_and [] false in
      let ls = read_list (next_non_ws tk_and) in
      node_p :: ls
   with Not_found ->
      let node_p = create_rec Proposition tk end_tk [] false in [node_p]
   in read_list tk

(** Proposition = [ Label-Identifier ":" ] Sentence . *)
let expand_proposition node =
  let tk = node.nt_first_tk in
  let tk1 = next_non_ws tk in
  if tk.tk_type = Miz_tk_ident && tk1.tk_type = Miz_colon then
    let node_label = create_rec Label_identifier tk tk1 [] true in
    let tk2 = next_non_ws tk1 in
    let node_s = create_rec Sentence tk2 node.nt_last_tk [] false in
    (node.nt_sublist <- [node_label; node_s]; node.nt_expanded <- true)
  else
    let node_s = create_rec Sentence tk node.nt_last_tk [] false in
    (node.nt_sublist <- [node_s]; node.nt_expanded <- true)

let expand_one node t =
  let n = create_rec t node.nt_first_tk node.nt_last_tk [] false in
  (node.nt_sublist <- [n]; node.nt_expanded <- true)

let is_mode_pattern tk_start =
  match tk_start.tk_type with
  | Miz_tk_symbol q when q = Voc.Mode -> true
  | Miz_set -> true
  | _ -> false

let is_attribute_pattern tk_start = (next_non_ws tk_start).tk_type = Miz_is

let is_predicate_pattern tk_start tk_end =
  let rec find tk =
    if tk == tk_end then false
    else
      match tk.tk_type with
      | Miz_tk_symbol q when q = Voc.Predicate -> true
      | Miz_eq -> true
      | _ -> find (next_non_ws tk)
  in find tk_start

let is_functor_pattern tk_start tk_end =
  let rec find tk =
    if tk == tk_end then false
    else
      match tk.tk_type with
      | Miz_tk_symbol q when (match q  with Voc.Functor _ -> true | _ -> false) -> true
      | _ -> find (next_non_ws tk)
  in match tk_start.tk_type with
  | Miz_tk_symbol q when q = Voc.Left_functor_bracket -> true
  | Miz_lc | Miz_lb -> true
  | _ -> find tk_start

let determine_kind tk_start tk_end =
  if is_attribute_pattern tk_start then Attribute_pattern
  else if is_mode_pattern tk_start then Mode_pattern
  else if is_predicate_pattern tk_start tk_end then Predicate_pattern
  else if is_functor_pattern tk_start tk_end then Functor_pattern
  else syntax_error_tk tk_start "Could not determine the kind of pattern"

(** Notation-Declaration = Attribute-Synonym | Attribute-Antonym | Functor-Synonym | Mode-Synonym | Predicate-Synonym | Predicate-Antonym . *)
let expand_notational_declaration node =
  let tk = node.nt_first_tk in
  let tk_semi = prev_non_ws node.nt_last_tk in
  let () = if tk_semi.tk_type <> Miz_semicolon then syntax_error_tk tk_semi "Expected ';'" in
  let tk_for = try find_first tk Miz_for tk_semi with Not_found -> syntax_error_tk tk "Expected for" in
  let t = determine_kind (next_non_ws tk_for) tk_semi in
  match t with
  | Attribute_pattern when tk.tk_type = Miz_synonym -> expand_one node Attribute_synonym
  | Attribute_pattern when tk.tk_type = Miz_antonym -> expand_one node Attribute_antonym
  | Mode_pattern when tk.tk_type = Miz_synonym -> expand_one node Mode_synonym
  | Predicate_pattern when tk.tk_type = Miz_synonym -> expand_one node Predicate_synonym
  | Predicate_pattern when tk.tk_type = Miz_antonym -> expand_one node Predicate_antonym
  | Functor_pattern when tk.tk_type = Miz_synonym -> expand_one node Functor_synonym
  | _ -> syntax_error_tk tk "Expected notation declaration"

let l_read_attribute_or_predicate_pattern tk_start tk_end =
  if is_attribute_pattern tk_start then l_read_attribute_pattern ([], tk_start)
  else if is_predicate_pattern tk_start tk_end then l_read_predicate_pattern ([], tk_start)
  else syntax_error_tk tk_start "Expected attribute or predicate pattern"
  
(** Attribute-Synonym = "synonym" ( Attribute-Pattern | Predicate-Pattern ) "for" Attribute-Pattern ";" . *)
let expand_attribute_synonym node =
  let tk = node.nt_first_tk in
  let () = if tk.tk_type <> Miz_synonym then syntax_error_tk tk "Expected synonym" in
  let tk_for = try find_first tk Miz_for node.nt_last_tk with Not_found -> syntax_error_tk tk "Expected for" in
  let (ls1, _) = l_read_attribute_or_predicate_pattern (next_non_ws tk) tk_for in
  let (ls2, _) = l_read_attribute_pattern ([], next_non_ws tk_for) in
  (node.nt_sublist <- ls1 @ ls2; node.nt_expanded <- true)

(** Attribute-Antonym = "antonym" ( Attribute-Pattern | Predicate-Pattern ) "for" Attribute-Pattern ";" . *)
let expand_attribute_antonym node =
  let tk = node.nt_first_tk in
  let () = if tk.tk_type <> Miz_antonym then syntax_error_tk tk "Expected antonym" in
  let tk_for = try find_first tk Miz_for node.nt_last_tk with Not_found -> syntax_error_tk tk "Expected for" in
  let (ls1, _) = l_read_attribute_or_predicate_pattern (next_non_ws tk) tk_for in
  let (ls2, _) = l_read_attribute_pattern ([], next_non_ws tk_for) in
  (node.nt_sublist <- ls1 @ ls2; node.nt_expanded <- true)

(** Functor-Synonym = "synonym" Functor-Pattern "for" Functor-Pattern ";" . *)
let expand_functor_synonym node =
  let tk = node.nt_first_tk in
  let () = if tk.tk_type <> Miz_synonym then syntax_error_tk tk "Expected synonym" in
  let (ls1, tk_for) = l_read_functor_pattern ([], next_non_ws tk) in
  let () = if tk_for.tk_type <> Miz_for then syntax_error_tk tk_for "Expected for" in
  let (ls2, _) = l_read_functor_pattern ([], next_non_ws tk_for) in
  (node.nt_sublist <- ls1 @ ls2; node.nt_expanded <- true)

(** Moder-Synonym = "synonym" Mode-Pattern "for" Mode-Pattern ";" . *)
let expand_mode_synonym node =
  let tk = node.nt_first_tk in
  let () = if tk.tk_type <> Miz_synonym then syntax_error_tk tk "Expected synonym" in
  let (n1, tk_for) = read_mode_pattern (next_non_ws tk) in
  let () = if tk_for.tk_type <> Miz_for then syntax_error_tk tk_for "Expected for" in
  let (n2, _) = read_mode_pattern (next_non_ws tk_for) in
  (node.nt_sublist <- [n1; n2]; node.nt_expanded <- true)

(** Predicate-Synonym = "synonym" Predicate-Pattern "for" Predicate-Pattern ";" . *)
let expand_predicate_synonym node =
  let tk = node.nt_first_tk in
  let () = if tk.tk_type <> Miz_synonym then syntax_error_tk tk "Expected synonym" in
  let (ls1, tk_for) = l_read_predicate_pattern ([], next_non_ws tk) in
  let () = if tk_for.tk_type <> Miz_for then syntax_error_tk tk_for "Expected for" in
  let (ls2, _) = l_read_predicate_pattern ([], next_non_ws tk_for) in
  (node.nt_sublist <- ls1 @ ls2; node.nt_expanded <- true)

(** Predicate-Antonym- = "antonym" Predicate-Pattern "for" Predicate-Pattern ";" . *)
let expand_predicate_antonym node =
  let tk = node.nt_first_tk in
  let () = if tk.tk_type <> Miz_antonym then syntax_error_tk tk "Expected antonym" in
  let (ls1, tk_for) = l_read_predicate_pattern ([], next_non_ws tk) in
  let () = if tk_for.tk_type <> Miz_for then syntax_error_tk tk_for "Expected for" in
  let (ls2, _) = l_read_predicate_pattern ([], next_non_ws tk_for) in
  (node.nt_sublist <- ls1 @ ls2; node.nt_expanded <- true)

(** Definiens = Simple-Definiens | Conditional-Definiens . *)
let expand_definiens node t =
  try
    let _ = find_first node.nt_first_tk Miz_if node.nt_last_tk in
    expand_one node (Simple_definiens t)
  with Not_found -> expand_one node (Conditional_definiens t)

(** Label-Identifier ":" *)
let read_label tk =
  let () = if tk.tk_type <> Miz_tk_ident then syntax_error_tk tk "Expected label identifier" in
  let tk1 = next_non_ws tk in
  let () = if tk1.tk_type <> Miz_colon then syntax_error_tk tk1 "Expected ':'" in
  (create_rec Label_identifier tk tk1 [] true, next_non_ws tk1)

(** [ Label-Identifier ":" ] *)
let read_label_opt tk =
  let tk1 = next_non_ws tk in
  if tk.tk_type = Miz_tk_ident && tk1.tk_type = Miz_colon then ([create_rec Label_identifier tk tk1 [] true], next_non_ws tk1)
  else ([], tk)

(** Simple-Definiens = [ ":" Label-Identifier ":" ] ( Sentence | Term-Expression ) . *)
let expand_simple_definiens node t =
  let read_opt_label tk = if tk.tk_type = Miz_colon then let (n, tk1) = read_label (next_non_ws tk) in ([n], tk1) else ([], tk) in
  let (ls, tk1) = read_opt_label node.nt_first_tk in
  let n = create_rec t tk1 node.nt_last_tk [] false in
  (node.nt_sublist <- ls  @ [n]; node.nt_expanded <- true)

(** Conditional-Definiens = [ ":" Label-Identifier ":" ] Partial-Definiens-List [ "otherwise" ( Sentence | Term-Expression ) ] . *)
let expand_conditional_definiens node t =
  let read_opt_rest ls tk =
    if tk.tk_type = Miz_otherwise then
      let n = create_rec t (next_non_ws tk) node.nt_last_tk [] false in ls @ [n]
   else []
  in
  let read_pd ls tk =
    let tk_next = try find_first tk Miz_otherwise node.nt_last_tk with Not_found -> node.nt_last_tk in
    let n = create_rec (Partial_definiens_list t) tk tk_next [] false in
    read_opt_rest (ls @ [n]) tk_next
  in
  let read tk = if tk.tk_type = Miz_colon then let (n, tk1) = read_label (next_non_ws tk) in read_pd [n] tk1 else read_pd [] tk in
  let ls = read node.nt_first_tk in
  (node.nt_sublist <- ls; node.nt_expanded <- true)

(** Partial-Definiens-List = Partial-Definiens { "," Partial-Definiens } . *)

(** Auxiliary-Item = Statement | Private-Definition . *)
let expand_auxiliary_item node =
  match node.nt_first_tk.tk_type with
  | Miz_set | Miz_deffunc | Miz_defpred -> expand_one node Private_definition
  | _ -> expand_one node Statement

(** Constant-Definition = "set" Equating-List ";" . *)
let expand_constant_definition node =
  let tk = node.nt_first_tk in
  let () = if tk.tk_type <> Miz_set then syntax_error_tk tk "Expected set" in
  let tk_e = next_non_ws tk in
  let tk_semi = prev_non_ws node.nt_last_tk in
  let n = create_rec Equating_list tk_e tk_semi [] false in
  (node.nt_sublist <- [n]; node.nt_expanded <- true)

(** Equating-List = Equating { "," Equating } . *)
(* Search for "," ident "=" which is outermost according to '{' '}' *)
let expand_equating_list node =
  let is_start_eq tk =
    let tk1 = next_non_ws tk in
    tk.tk_type = Miz_tk_ident && tk1.tk_type = Miz_eq
  in
  let rec find_next tk level =
    if tk == node.nt_last_tk then raise Not_found else
      let tk1 = next_non_ws tk in
      match tk.tk_type with
      | Miz_lc -> find_next tk1 (level+1)
      | Miz_rc -> find_next tk1 (level-1)
      | Miz_comma when level = 0 -> if is_start_eq tk1 then tk else find_next tk1 level
      | _ -> find_next tk1 level
  in
  let rec read_list tk =
    try
      let tk1 = find_next tk 0 in
      let n  = create_rec Equating tk tk1 [] false in
      let (u, v) = read_list (next_non_ws tk1) in
      (n :: u, v)
    with Not_found -> ([create_rec Equating tk node.nt_last_tk [] false], node.nt_last_tk)
  in
  let tk = node.nt_first_tk in
  let (ls, _) = read_list tk in
  (node.nt_sublist <- ls; node.nt_expanded <- true)

(** Equating = Variable-Identifier "=" Term-Expression . *)
let expand_equating node =
  let tk = node.nt_first_tk in  
  let () = if tk.tk_type <> Miz_tk_ident then syntax_error_tk tk "Expected variable name" in
  let tk_eq = next_non_ws tk in
  let () = if tk_eq.tk_type <> Miz_eq then syntax_error_tk tk "Expected =" in
  let n_v = read_variable_identifier tk in
  let n_t = create_rec Term_expression (next_non_ws tk_eq) node.nt_last_tk [] false in
  (node.nt_sublist <- [n_v; n_t]; node.nt_expanded <- true)

(** Private-Functor-Definition = "deffunc" Private-Functor-Pattern "=" Term-Expression . *)
let expand_private_functor_definition node =
  let tk = node.nt_first_tk in
  let () = if tk.tk_type <> Miz_deffunc then syntax_error_tk tk "Expected deffunc" in
  let tk = next_non_ws tk in
  try
    let tk_eq = find_outer_lc_rc tk node.nt_last_tk Miz_eq in
    let n1 = create_rec Private_functor_pattern tk tk_eq [] false in
    let n2 = create_rec Term_expression (next_non_ws tk_eq) node.nt_last_tk [] false in
    (node.nt_sublist <- [n1; n2]; node.nt_expanded <- true)
  with Not_found -> syntax_error_tk tk "Not found = in private functor definition"

(** Private-Predicate-Definition = "defpred" Private-Predicate-Pattern "means" Sentence . *)
let expand_private_predicate_definition node =
  let tk = node.nt_first_tk in
  let () = if tk.tk_type <> Miz_defpred then syntax_error_tk tk "Expected defpred" in
  let tk = next_non_ws tk in
  try
    let tk_means = find_outer_lc_rc tk node.nt_last_tk Miz_means in
    let n1 = create_rec Private_predicate_pattern tk tk_means [] false in
    let n2 = create_rec Sentence (next_non_ws tk_means) node.nt_last_tk [] false in
    (node.nt_sublist <- [n1; n2]; node.nt_expanded <- true)
  with Not_found -> syntax_error_tk tk "Not found means in private predicate definition"

(** Private-Functor-Pattern = Functor-Identifier "(" [ Type-Expression-List ] ")" . *)
let expand_private_functor_pattern node =
  let tk = node.nt_first_tk in
  let () = if tk.tk_type <> Miz_tk_ident then syntax_error_tk tk "Expected private functor name" in
  let tk_lp = next_non_ws tk in
  let () = if tk_lp.tk_type <> Miz_lp then syntax_error_tk tk_lp "Expected (" in
  let n1 = create_rec Functor_identifier tk tk_lp [] true in
  let tk_rp = prev_non_ws node.nt_last_tk in
  let () = if tk_rp.tk_type <> Miz_rp then syntax_error_tk tk_rp "Expected )" in
  let n2 = create_rec Type_expression_list (next_non_ws tk_lp) tk_rp [] false in
  (node.nt_sublist <- [n1; n2]; node.nt_expanded <- true)

(** Private-Predicate-Pattern = Predicate-Identifier "[" [ Type-Expression-List ] "]" . *)
let expand_private_predicate_pattern node =
  let tk = node.nt_first_tk in
  let () = if tk.tk_type <> Miz_tk_ident then syntax_error_tk tk "Expected private predicate name" in
  let tk_lb = next_non_ws tk in
  let () = if tk_lb.tk_type <> Miz_lb then syntax_error_tk tk_lb "Expected [" in
  let n1 = create_rec Predicate_identifier tk tk_lb [] true in
  let tk_rb = prev_non_ws node.nt_last_tk in
  let () = if tk_rb.tk_type <> Miz_rb then syntax_error_tk tk_rb "Expected ]" in
  let n2 = create_rec Type_expression_list (next_non_ws tk_lb) tk_rb [] false in
  (node.nt_sublist <- [n1; n2]; node.nt_expanded <- true)

(** Statement = "then" ] Linkable-Statement | Diffuse-Statement . *)
let expand_statement node =
  let is_diffuse tk =
    tk.tk_type = Miz_now || (next_non_ws (next_non_ws tk)).tk_type = Miz_now in
  let tk = node.nt_first_tk in
  if is_diffuse tk then expand_one node Diffuse_statement else
    let tk = if tk.tk_type = Miz_then then next_non_ws tk else tk in
    let n = create_rec Linkable_statement tk node.nt_last_tk [] false in
    (node.nt_sublist <- [n]; node.nt_expanded <- true)

(**Linkable-Statement = Compact-Statement | Choice-Statement | Type-Changing-Statement | Iterative-Equality . *) 
let expand_linkable_statement node =
  let is_it_eq () =
    try
      let _ = find_outer node.nt_first_tk node.nt_last_tk Miz_seq_eq
      in true
    with Not_found -> false
  in
  let tk = node.nt_first_tk in
  match tk.tk_type with
  | Miz_consider -> expand_one node Choice_statement
  | Miz_reconsider -> expand_one node Type_changing_statement
  | _ when is_it_eq () -> expand_one node Iterative_equality
  | _ -> expand_one node Compact_statement

(** Iterative-Equality = [ Label-Identifier  ":" ] Term-Expression "=" Term-Expression Simple-Justification ".=" Term-Expression Simple-Justification { ".=" Term-Expression Simple-Justification } ";" . *)
let expand_iterative_equality node =
  let read_item tk =
    let tk_next = try find_outers tk node.nt_last_tk [Miz_by; Miz_from; Miz_seq_eq; Miz_semicolon] with Not_found -> syntax_error_tk tk "There is no ending ';'" in
    let n1 = create_rec Term_expression tk tk_next [] false in
    let (n2, tk_next) = read_simple_justification tk_next in
    ([n1; n2], tk_next)
  in
  let rec read_list tk =
    let (ls1, tk1) = read_item tk in
    match tk1.tk_type with
    | Miz_seq_eq -> let (u, v) = read_list (next_non_ws tk1) in (ls1 @ u, v)
    | Miz_semicolon -> (ls1, tk1)
    | _ -> syntax_error_tk tk1 "Expected '.=' or ';'"
  in
  let tk = node.nt_first_tk in
  let tk_eq = try find_outer node.nt_first_tk node.nt_last_tk Miz_eq with Not_found -> syntax_error_tk tk "There is no '='" in
  let (ls1, tk1) = read_label_opt tk in
  let n1 = create_rec Term_expression tk1 tk_eq [] false in
  (* let () = if (List.length ls1 > 0) then print_token tk in *)
  let (ls, _) = read_list (next_non_ws tk_eq) in
  (node.nt_sublist <- ls1 @ [n1] @ ls; node.nt_expanded <- true)

(** Type-Changing-Statement = "reconsider" Type-Change-List "as" Type-Expression Simple-Justification ";" . *)
let expand_type_changing_statement node =
  let tk = node.nt_first_tk in
  let () = if tk.tk_type <> Miz_reconsider then syntax_error_tk tk "Expected reconsider" in
  let tk_as = try find_first tk Miz_as node.nt_last_tk with Not_found -> syntax_error_tk tk "There is no as" in
  let n1 = create_rec Type_change_list (next_non_ws tk) tk_as [] false in
  let tk_semi = prev_non_ws node.nt_last_tk in
  let () = if tk_semi.tk_type <> Miz_semicolon then syntax_error_tk tk_semi "Expected ';'" in
  let tk1 = try find_outers tk tk_semi [Miz_by; Miz_from] with Not_found -> tk_semi in
  let n2 = create_rec Type_expression (next_non_ws tk_as) tk1 [] false in
  let n3 = create_rec Simple_justification tk1 tk_semi [] false in
  (node.nt_sublist <- [n1; n2; n3]; node.nt_expanded <- true)

(** Private-Definition = Constant-Definition | Private-Functor-Definition | Private-Predicate-Definition . *)
let expand_private_definition node =
  let tk = node.nt_first_tk in
  match tk.tk_type with
  | Miz_set -> expand_one node Constant_definition
  | Miz_deffunc -> expand_one node Private_functor_definition
  | Miz_defpred -> expand_one node Private_predicate_definition
  | _ -> syntax_error_tk tk "Expected set, deffunc or defpred"

(** Theorem = "theorem" Compact-Statement . *)
let expand_theorem node =
  let tk = node.nt_first_tk in
  let () = if tk.tk_type <> Miz_theorem then syntax_error_tk tk "Expected theorem" in
  let tk = next_non_ws tk in
  let n = create_rec Compact_statement tk node.nt_last_tk [] false in
  (node.nt_sublist <- [n]; node.nt_expanded <- true)

(** Compact-Statement = Proposition Justification ";" . *)
let expand_compact_statement node =
  let tk = node.nt_first_tk in
  let tk_semi = prev_non_ws node.nt_last_tk in
  let () = if tk_semi.tk_type <> Miz_semicolon then syntax_error_tk tk_semi "Expected ';'" in
  try
    let tk_j = find_firsts tk [Miz_by; Miz_from; Miz_proof; Miz_at_proof] tk_semi in
    let n1 = create_rec Proposition tk tk_j [] false in
    let n2 = create_rec Justification tk_j tk_semi [] false in
    (node.nt_sublist <- [n1; n2]; node.nt_expanded <- true)
  with Not_found -> let n1 = create_rec Proposition tk tk_semi [] false in
    let n2 = create_rec Justification tk_semi tk_semi [] false in
    (node.nt_sublist <- [n1; n2]; node.nt_expanded <- true)

(** Justification = Simple-Justification | Proof . *)
let expand_justification node =
  let tk = node.nt_first_tk in
  if tk == node.nt_last_tk then expand_one node Simple_justification else
    match tk.tk_type with (* TODO *)
    | Miz_by | Miz_from -> let (n, _) = read_simple_justification tk in (node.nt_sublist <- [n]; node.nt_expanded <- true)
    | Miz_proof | Miz_at_proof -> expand_one node Proof
    | _ -> syntax_error_tk tk "Expected by, from, proof or @proof"

(** Simple-Justification = Straightforward-Justification | Scheme-Justification . *)
let expand_simple_justification node =
  let tk = node.nt_first_tk in
  let (n, _) = read_simple_justification tk in (node.nt_sublist <- [n]; node.nt_expanded <- true)
(*
  let tk = node.nt_first_tk in
  if tk == node.nt_last_tk then expand_one node Straightforward_justification else
    match tk.tk_type with
    | Miz_by -> expand_one node Straightforward_justification
    | Miz_from -> expand_one node Scheme_justification
    | _ -> syntax_error_tk tk "Expected by or from"
*)

(** Scheme-Justification = "from" Scheme-Reference [ "(" References ")" ] . *)
(* let expand_scheme_justification node = failwith "expand_scheme_justification not impemented" *)
  (* let tk = node.nt_first_tk in *)
  (* let () = if tk.nt_first_tk <> Miz_from then syntax_error_tk tk "Expected from" in *)

(** Proof = ( "proof" | "@proof" ) Reasoning "end" . *)
let expand_proof node =
  let tk = node.nt_first_tk in
  let () = if tk.tk_type <> Miz_proof && tk.tk_type <> Miz_at_proof then syntax_error_tk tk "Expected proof or @proof" in
  let tk_end = prev_non_ws node.nt_last_tk in
  let () = if tk_end.tk_type <> Miz_end then syntax_error_tk tk_end "Expected end" in
  let n = create_rec Reasoning (next_non_ws tk) tk_end [] false in
  (node.nt_sublist <- [n]; node.nt_expanded <- true)

let read_reasoning_item_list tk_start tk_end =
  (* let () = begin print_string "--reasoning_item_list start "; print_token tk_start; print_string "--reasoning_item_list end "; print_token tk_end end in *)
  let rec read_list tk =
    if tk == tk_end then [] else
    try
      let tk_semi = find_outer tk tk_end Miz_semicolon in
      (* let () = begin print_string "semi "; print_token tk_semi end in *)
      let tk_next = next_non_ws tk_semi in
      let n = create_rec Reasoning_item tk tk_next [] false in
      let ls = read_list tk_next in
      n :: ls
    with Not_found -> begin print_token tk_start; print_token tk_end; syntax_error_tk tk_end "Expected ';'" end
  in read_list tk_start

(** Reasoning = { Reasoning-Item } [ "per" "cases" Simple-Justification ";" ( Case-List | Suppose-List ) ] . *)
let expand_reasoning node =
  (* let () = begin print_string "--reasoning start node.nt_first_tk "; print_token node.nt_first_tk; print_string "--reasoning end node.nt_last_tk "; print_token node.nt_last_tk end in *)
  let tk = node.nt_first_tk in
  try
    let tk_per = find_outer tk node.nt_last_tk Miz_per in
    let ls = read_reasoning_item_list tk tk_per in
    let tk_cases = next_non_ws tk_per in
    let () = if tk_cases.tk_type <> Miz_cases then syntax_error_tk tk_cases "Expected cases" in
    let tk = next_non_ws tk_cases in
    let tk_semi =
      try
        find_outer tk node.nt_last_tk Miz_semicolon
      with Not_found -> syntax_error_tk (prev_non_ws node.nt_last_tk) "Expected ';'" in
    let n1 = create_rec Simple_justification tk tk_semi [] false in
    let tk = next_non_ws tk_semi in
    let n2 =
      match tk.tk_type with
      | Miz_case -> create_rec Case_list tk node.nt_last_tk [] false
      | Miz_suppose -> create_rec Suppose_list tk node.nt_last_tk [] false
      | _ -> syntax_error_tk tk "Expected case or suppose"
    in
    (node.nt_sublist <- ls @ [n1; n2]; node.nt_expanded <- true)
  with Not_found ->
    (* let () = begin print_endline "not found start list"; print_token tk; print_endline "not found node.nt_last "; print_token node.nt_last_tk end in *)
    let ls = read_reasoning_item_list tk node.nt_last_tk in
    (node.nt_sublist <- ls; node.nt_expanded <- true)

(** Case-List = Case { Case } . *)
let expand_case_list node =
  let rec read_list tk tk_end =
    if tk == tk_end then [] else
      if tk.tk_type = Miz_case then
        let tk_semi =
          try
            find_outer tk node.nt_last_tk Miz_semicolon
          with Not_found -> syntax_error_tk tk "There is no matching end" in
        let tk_next = next_non_ws tk_semi in
        let n = create_rec Case tk tk_next [] false in
        let ls = read_list tk_next tk_end in
        n :: ls
      else syntax_error_tk tk "Expected case"
  in
  let tk = node.nt_first_tk in
  let () = if tk.tk_type <> Miz_case then syntax_error_tk tk "Expected case" in
  let ls = read_list tk node.nt_last_tk in
  (node.nt_sublist <- ls; node.nt_expanded <- true)

(** Suppose-List = Suppose { Suppose } . *)
let expand_suppose_list node =
  (* let () = begin print_string "--suppose_list start node.nt_first_tk "; print_token node.nt_first_tk; print_string "--suppose_list end node.nt_last_tk "; print_token node.nt_last_tk end in *)
  let rec read_list tk tk_end =
    if tk == tk_end then [] else
      if tk.tk_type = Miz_suppose then
        let tk_semi =
          try
            find_outer tk node.nt_last_tk Miz_semicolon
          with Not_found -> syntax_error_tk tk "There is no matching end" in
        let tk_next = next_non_ws tk_semi in
        (* let () = begin print_string "--suppose start "; print_token tk; print_string "--suppose end "; print_token tk_next end in *)
        let n = create_rec Suppose tk tk_next [] false in
        let ls = read_list tk_next tk_end in
        n :: ls
      else syntax_error_tk tk "Expected suppose"
  in
  let tk = node.nt_first_tk in
  let () = if tk.tk_type <> Miz_suppose then syntax_error_tk tk "Expected suppose" in
  let ls = read_list tk node.nt_last_tk in
  (node.nt_sublist <- ls; node.nt_expanded <- true)

(** Case = "case" ( Proposition | Conditions ) ";" Reasoning "end" ";" . *)
let expand_case node =
  let tk = node.nt_first_tk in
  let () = if tk.tk_type <> Miz_case then syntax_error_tk tk "Expected case" in
  let tk = next_non_ws tk in
  let tk_semi =
    try find_outer tk node.nt_last_tk Miz_semicolon with Not_found -> syntax_error_tk tk "There is no delimiter ';'" in
  let n =
    match tk.tk_type with
    | Miz_that -> create_rec Conditions tk tk_semi [] false
    | _ -> create_rec Proposition tk tk_semi [] false
  in
  let tk_last = prev_non_ws node.nt_last_tk in
  let () = if tk_last.tk_type <> Miz_semicolon then syntax_error_tk tk_last "Expected ';'" in
  let tk_end = prev_non_ws tk_last in
  let () = if tk_end.tk_type <> Miz_end then syntax_error_tk tk_last "Expected end" in
  let n1 = create_rec Reasoning (next_non_ws tk_semi) tk_end [] false in
  (node.nt_sublist <- [n; n1]; node.nt_expanded <- true)

(** Suppose = "suppose" ( Proposition | Conditions ) ";" Reasoning  "end" ";" . *)
let expand_suppose node =
  (* let () = begin print_string "--suppose start node.nt_first_tk "; print_token node.nt_first_tk; print_string "--suppose end node.nt_last_tk "; print_token node.nt_last_tk end in *)
  let tk = node.nt_first_tk in
  let () = if tk.tk_type <> Miz_suppose then syntax_error_tk tk "Expected suppose" in
  let tk = next_non_ws tk in
  let tk_semi =
    try find_outer tk node.nt_last_tk Miz_semicolon with Not_found -> syntax_error_tk tk "There is no delimiter ';'" in
  (* let () = begin print_string "tk_semi "; print_token tk_semi end in *)
  let n =
    match tk.tk_type with
    | Miz_that -> create_rec Conditions tk tk_semi [] false
    | _ -> create_rec Proposition tk tk_semi [] false
  in
  let tk_last = prev_non_ws node.nt_last_tk in
  let () = if tk_last.tk_type <> Miz_semicolon then syntax_error_tk tk_last "Expected ';'" in
  let tk_end = prev_non_ws tk_last in
  let () = if tk_end.tk_type <> Miz_end then syntax_error_tk tk_last "Expected end" in
  let n1 = create_rec Reasoning (next_non_ws tk_semi) tk_end [] false in
  (node.nt_sublist <- [n; n1]; node.nt_expanded <- true)


(** Reasoning-Item = Auxiliary-Item | Skeleton-Item . *)
let expand_reasoning_item node =
  match node.nt_first_tk.tk_type with
  | Miz_let | Miz_assume | Miz_given | Miz_thus | Miz_hence | Miz_hereby | Miz_take -> expand_one node Skeleton_item
  | _ -> expand_one node Auxiliary_item

(** Skeleton-Item = Generalization | Assumption | Conclusion | Exemplification . *)
let expand_skeleton_item node =
  match node.nt_first_tk.tk_type with
  | Miz_let -> expand_one node Generalization
  | Miz_assume | Miz_given -> expand_one node Assumption
  | Miz_thus | Miz_hence | Miz_hereby -> expand_one node Conclusion
  | Miz_take -> expand_one node Exemplification
  | _ -> syntax_error_tk node.nt_first_tk "Expected Skeleton-Item"

(** Single-Assumption = "assume" Proposition ";" . *)
let expand_single_assumption node =
  let tk = node.nt_first_tk in
  let () = if tk.tk_type <> Miz_assume then syntax_error_tk tk "Expected assume" in
  let tk_p = next_non_ws tk in
  let tk_semi = prev_non_ws node.nt_last_tk in
  let n = create_rec Proposition tk_p tk_semi [] false in
  (node.nt_sublist <- [n]; node.nt_expanded <- true)

(** Collective-Assumption = "assume" Conditions ";" . *)
let expand_collective_assumption node =
  let tk = node.nt_first_tk in
  let () = if tk.tk_type <> Miz_assume then syntax_error_tk tk "Expected assume" in
  let tk_p = next_non_ws tk in
  let tk_semi = prev_non_ws node.nt_last_tk in
  let n = create_rec Conditions tk_p tk_semi [] false in
  (node.nt_sublist <- [n]; node.nt_expanded <- true)

(** Existential-Assumption = "given" Qualified-Variables [ "such" Conditions ] ";" . *)
let expand_existential_assumption node =
  let tk = node.nt_first_tk in
  let () = if tk.tk_type <> Miz_given then syntax_error_tk tk "Expected given" in
  let tk_q = next_non_ws tk in
  let tk_semi = prev_non_ws node.nt_last_tk in
  try
    let tk_such = find_first tk Miz_such tk_semi in
    let node_q = create_rec Qualified_variables tk_q tk_such [] false in
    let node_c = create_rec Conditions (next_non_ws tk_such) tk_semi [] false in
    (node.nt_sublist <- [node_q; node_c]; node.nt_expanded <- true)
  with Not_found -> let node_q = create_rec Qualified_variables tk_q tk_semi [] false in
    (node.nt_sublist <- [node_q]; node.nt_expanded <- true)

(** Assumption = Single-Assumption | Collective-Assumption | Existential-Assumption . *)
let expand_assumption node =
  let tk = node.nt_first_tk in
  let tk1 = next_non_ws tk in
  match tk.tk_type with
  | Miz_assume -> expand_one node (if tk1.tk_type = Miz_that then Collective_assumption else Single_assumption)
  | Miz_given -> expand_one node Existential_assumption
  | _ -> syntax_error_tk tk "Expected assume or given"

(** Loci-Declaration = "let" Qualified-Variables [ "such" Conditions ] ";" . *)
let expand_loci_declaration node =
  let tk = node.nt_first_tk in
  let () = if tk.tk_type <> Miz_let then syntax_error_tk tk "Expected let" in
  let tk_q = next_non_ws tk in
  let tk_semi = prev_non_ws node.nt_last_tk in
  try
    let tk_such = find_first tk Miz_such tk_semi in
    let node_q = create_rec Qualified_variables tk_q tk_such [] false in
    let node_c = create_rec Conditions (next_non_ws tk_such) tk_semi [] false in
    (node.nt_sublist <- [node_q; node_c]; node.nt_expanded <- true)
  with Not_found -> let node_q = create_rec Qualified_variables tk_q tk_semi [] false in
    (node.nt_sublist <- [node_q]; node.nt_expanded <- true)

(** Generalization = "let" Qualified-Variables [ "such" Conditions ] ";" . *)
let expand_generalization node =
  let tk = node.nt_first_tk in
  let () = if tk.tk_type <> Miz_let then syntax_error_tk tk "Expected let" in
  let tk_q = next_non_ws tk in
  let tk_semi = prev_non_ws node.nt_last_tk in
  try
    let tk_such = find_first tk Miz_such tk_semi in
    let node_q = create_rec Qualified_variables tk_q tk_such [] false in
    let node_c = create_rec Conditions (next_non_ws tk_such) tk_semi [] false in
    (node.nt_sublist <- [node_q; node_c]; node.nt_expanded <- true)
  with Not_found -> let node_q = create_rec Qualified_variables tk_q tk_semi [] false in
    (node.nt_sublist <- [node_q]; node.nt_expanded <- true)

(* Simple-Justification ";"  = "by" ... | "from" ... | ";" ... *)
(** Choice-Statement = "consider" Qualified-Variables [ "such" Conditions ] Simple-Justification ";" . *)
let expand_choice_statement node =
  let tk = node.nt_first_tk in
  let () = if tk.tk_type <> Miz_consider then syntax_error_tk tk "Expected consider" in
  let tk_q = next_non_ws tk in  
  let tk_semi = prev_non_ws node.nt_last_tk in
  let () = if tk_semi .tk_type <> Miz_semicolon then syntax_error_tk tk_semi "Expected ';'" in
  let tk_j = try find_firsts tk_q [Miz_by; Miz_from] tk_semi with Not_found -> tk_semi in
  try
    let tk_such = find_first tk Miz_such tk_semi in
    let node_q = create_rec Qualified_variables tk_q tk_such [] false in
    let node_c = create_rec Conditions (next_non_ws tk_such) tk_j [] false in
    let node_j = create_rec Simple_justification tk_j tk_semi [] false in
    (node.nt_sublist <- [node_q; node_c; node_j]; node.nt_expanded <- true)
  with Not_found ->
    let node_q = create_rec Qualified_variables tk_q tk_j [] false in
    let node_j = create_rec Simple_justification tk_j tk_semi [] false in
    (node.nt_sublist <- [node_q; node_j]; node.nt_expanded <- true)

(** Conclusion = ( "thus" | "hence" ) Compact-Statement | Diffuse-Conclusion . *)
let expand_conclusion node =
  let is_diffuse_statement tk =
    let tk1 = next_non_ws tk in
    let tk2 = next_non_ws tk1 in
    tk.tk_type = Miz_now || (tk.tk_type = Miz_tk_ident && tk1.tk_type = Miz_colon && tk2.tk_type = Miz_now)
  in
  let tk = node.nt_first_tk in
  let tk1 = next_non_ws tk in
  match tk.tk_type with
  | Miz_hence -> let n = create_rec Compact_statement tk1 node.nt_last_tk [] false in
                 (node.nt_sublist <- [n]; node.nt_expanded <- true)
  | Miz_thus when is_diffuse_statement tk1 -> expand_one node Diffuse_conclusion
  | Miz_hereby -> expand_one node Diffuse_conclusion
  | Miz_thus -> let n = create_rec Compact_statement tk1 node.nt_last_tk [] false in
                 (node.nt_sublist <- [n]; node.nt_expanded <- true)
  | _ -> syntax_error_tk tk "Expected thus, hence or hereby"

(** Diffuse-Conclusion = "thus" Diffuse-Statement | "hereby" Reasoning "end" ";" . *)
let expand_diffuse_conclusion node =
  let tk = node.nt_first_tk in
  let tk1 = next_non_ws tk in
  match tk.tk_type with
  | Miz_thus -> let n = create_rec Diffuse_statement tk1 node.nt_last_tk [] false in
                (node.nt_sublist <- [n]; node.nt_expanded <- true)
  | Miz_hereby -> let tk_semi = prev_non_ws node.nt_last_tk in
                  let () = if tk_semi.tk_type <> Miz_semicolon then syntax_error_tk tk_semi "Expected ';'" in
                  let tk_end = prev_non_ws tk_semi in
                  let () = if tk_end.tk_type <> Miz_end then syntax_error_tk tk_end "Expected end" in
                  let n = create_rec Reasoning tk1 tk_end [] false in
                  (node.nt_sublist <- [n]; node.nt_expanded <- true)
  | _ -> syntax_error_tk tk "Expected thus or hereby"

(** Diffuse-Statement = [ Label-Identifier ":" ] "now" Reasoning "end" ";" . *)
let expand_diffuse_statement node =
  let read_now tk =
    let () = if tk.tk_type <> Miz_now then syntax_error_tk tk "Expected now" in
    let tk1 = next_non_ws tk in
    let tk_semi = prev_non_ws node.nt_last_tk in
    let () = if tk_semi.tk_type <> Miz_semicolon then syntax_error_tk tk_semi "Expected ';'" in
    let tk_end = prev_non_ws tk_semi in
    let () = if tk_end.tk_type <> Miz_end then syntax_error_tk tk_end "Expected end" in
    let n = create_rec Reasoning tk1 tk_end [] false in
    [n]
  in
  let read_list tk =
    if tk.tk_type = Miz_tk_ident then
      let tk_colon = next_non_ws tk in
      let () = if tk_colon.tk_type <> Miz_colon then syntax_error_tk tk_colon "Expected ':'" in
      let tk1 = next_non_ws tk_colon in    
      let n = create_rec Label_identifier tk tk_colon [] true in
      n :: (read_now tk1)
    else
      read_now tk
  in
  let tk = node.nt_first_tk in
  let ls = read_list tk in
  (node.nt_sublist <- ls; node.nt_expanded <- true)

let read_structure_symbol tk =
  match tk.tk_type with
  | Miz_tk_symbol q when q = Voc.Structure ->
      let tk1 = next_non_ws tk in (create_rec Structure_symbol tk tk1 [] true, tk1)
  | _ -> syntax_error_tk tk "Expected structure symbol"

(** Structure-Definition = "struct" [ "(" Ancestors ")" ] Structure-Symbol [ "over" Loci ] "(#" Fields "#)" ";" . *)
let expand_structure_definition node =
  let read_ancestors tk =
    if tk.tk_type = Miz_lp then
      let tk_rp =
        try find_matching_rp tk node.nt_last_tk with Not_found -> syntax_error_tk tk "There is no matching ')'" in
      let n = create_rec Ancestors (next_non_ws tk) tk_rp [] false in
      ([n], next_non_ws tk_rp)
    else ([], tk)
  in
  let read_l tk =
    if tk.tk_type = Miz_over then
      let (loci_list, tk1) = read_loci ([], (next_non_ws tk)) in
      (loci_list, tk1)
    else ([], tk)
  in
  let tk = node.nt_first_tk in
  let () = if tk.tk_type <> Miz_struct then syntax_error_tk tk "Expected struct" in
  let tk = next_non_ws tk in
  let (ls_a, tk) = read_ancestors tk in
  let (n_s, tk) = read_structure_symbol tk in
  let (ls_l, tk) = read_l tk in
  let () = if tk.tk_type <> Miz_lpound then syntax_error_tk tk "Expected '(#'" in
  let tk = next_non_ws tk in
  let tk_semi = prev_non_ws node.nt_last_tk in
  let () = if tk_semi.tk_type <> Miz_semicolon then syntax_error_tk tk_semi "Expected ';'" in
  let tk_rpound = prev_non_ws tk_semi in
  let () = if tk_rpound.tk_type <> Miz_rpound then syntax_error_tk tk_rpound "Expected '#)'" in
  let n = create_rec Fields tk tk_rpound [] false in
  (node.nt_sublist <- ls_a @ [n_s] @ ls_l @ [n]; node.nt_expanded <- true)

(** Fields = Field-Segment { "," Field-Segment } . *)
let expand_fields node =
  (* let () = print_token node.nt_first_tk in *)
  (* let () = print_token node.nt_last_tk in *)
  (* let () = print_endline "---------------------" in *)
  let is_delimiter_semi tk =
    tk.tk_type = Miz_comma && (is_selector_tk (next_non_ws tk))
  in
  let rec read_up_to tk_start tk_end =
    if tk_start == tk_end then tk_end
    else
      if is_delimiter_semi tk_start then tk_start else read_up_to (next_non_ws tk_start) tk_end
  in
  let rec read_segments tk_start tk_end =
    if tk_start == tk_end then []
    else
      let tk_imp = try find_first tk_start Miz_imp tk_end with Not_found -> syntax_error_tk tk_start "There is no '->'" in
      let tk_next = read_up_to tk_imp tk_end in
      let n = create_rec Field_segment tk_start tk_next [] false in
      (* let () = print_endline (node_txt n) in *)
      if tk_next == tk_end then [n] else n :: (read_segments (next_non_ws tk_next) tk_end)
  in
  let ls = read_segments node.nt_first_tk node.nt_last_tk in
  (node.nt_sublist <- ls; node.nt_expanded <- true)  

(** Field-Segment = Selector-Symbol { "," Selector-Symbol } Specification . *)
let expand_field_segment node =
  let rec read_list tk =
    let (n, tk1) = read_selector_symbol tk in
    if tk1.tk_type = Miz_comma then let (u, v) = read_list (next_non_ws tk1) in (n :: u, v)
    else ([n], tk1)
  in
  let (ls, tk_imp) = read_list node.nt_first_tk in
  let () = if tk_imp.tk_type <> Miz_imp then syntax_error_tk tk_imp "Expected '->'" in
  let n = create_rec Specification tk_imp node.nt_last_tk [] false in
  (node.nt_sublist <- ls @ [n]; node.nt_expanded <- true)

(** Scheme-Conclusion = Sentence . *)
let expand_scheme_conclusion node = expand_one node Sentence

(** Scheme-Premise = Proposition . *)
let expand_scheme_premise node = expand_one node Proposition

(** Reservation = "reserve" Reservation-Segment { "," Reservation-Segment } ";" . *)
let expand_reservation node =
  let tk1 = find_outer_lc_rc node.nt_first_tk node.nt_last_tk Miz_for in
  try
    let _ = find_outer_lc_rc (next_non_ws tk1) node.nt_last_tk Miz_for in
    ()
    (* print_endline (node_txt node) *)  
  with Not_found -> ()

let expand_node node =
  if not node.nt_expanded then
  let one t =
    let n = create_rec t node.nt_first_tk node.nt_last_tk [] false in
    (node.nt_sublist <- [n]; node.nt_expanded <- true)
  in
  match node.nt_type with
  | Conditions -> let ls = expand_conditions node.nt_first_tk node.nt_last_tk in
                  (node.nt_sublist <- ls; node.nt_expanded <- true)
  | Proposition -> expand_proposition node
  | Sentence -> one Formula_expression
  | Permissive_assumption -> one Assumption
  | Assumption -> expand_assumption node
  | Single_assumption -> expand_single_assumption node
  | Collective_assumption -> expand_collective_assumption node
  | Existential_assumption -> expand_existential_assumption node
  | Loci_declaration -> expand_loci_declaration node
  | Auxiliary_item -> expand_auxiliary_item node
  | Private_definition -> expand_private_definition node
  | Constant_definition -> expand_constant_definition node
  | Equating_list -> expand_equating_list node
  | Equating -> expand_equating node
  | Private_functor_definition -> expand_private_functor_definition node
  | Private_predicate_definition -> expand_private_predicate_definition node
  | Private_functor_pattern -> expand_private_functor_pattern node
  | Private_predicate_pattern -> expand_private_predicate_pattern node
  | Statement -> expand_statement node
  | Linkable_statement -> expand_linkable_statement node
  | Theorem -> expand_theorem node
  | Compact_statement -> expand_compact_statement node
  | Justification -> expand_justification node
  | Simple_justification -> expand_simple_justification node
  | Proof -> expand_proof node
  | Reasoning -> expand_reasoning node
  | Case_list -> expand_case_list node
  | Suppose_list -> expand_suppose_list node
  | Case -> expand_case node
  | Suppose -> expand_suppose node
  | Reasoning_item -> expand_reasoning_item node
  | Skeleton_item -> expand_skeleton_item node
  | Generalization -> expand_generalization node
  | Conclusion -> expand_conclusion node
  | Diffuse_conclusion -> expand_diffuse_conclusion node
  | Diffuse_statement -> expand_diffuse_statement node
  | Structure_definition -> expand_structure_definition node
  | Specification -> expand_specification node
  | Definiens t -> expand_definiens node t
  | Simple_definiens t -> expand_simple_definiens node t
  | Conditional_definiens t -> expand_conditional_definiens node t
  | Notation_declaration -> expand_notational_declaration node
  | Attribute_synonym -> expand_attribute_synonym node
  | Attribute_antonym -> expand_attribute_antonym node
  | Functor_synonym -> expand_functor_synonym node
  | Mode_synonym -> expand_mode_synonym node
  | Predicate_synonym -> expand_predicate_synonym node
  | Predicate_antonym -> expand_predicate_antonym node
  | Scheme_conclusion -> expand_scheme_conclusion node
  | Scheme_premise -> expand_scheme_premise node
  | Choice_statement -> expand_choice_statement node
  | Fields -> expand_fields node
  | Field_segment -> expand_field_segment node
  | Ancestors -> expand_ancestors node
  | Iterative_equality -> expand_iterative_equality node
  | Type_changing_statement -> expand_type_changing_statement node
  | Reservation -> expand_reservation node
  | _ -> ()

let rec expand_tree node =
  let () = if not node.nt_expanded then expand_node node in
  if node.nt_expanded then
    List.iter expand_tree node.nt_sublist

let rec collect_non_expanded node =
  if not node.nt_expanded then let (name, _) = ast_miz_to_tag node.nt_type in [name]
  else List.concat (List.map collect_non_expanded node.nt_sublist)

(** Permissive-Assumption = Assumption . *)
let read_permissive_assumption tk =
  let tk1 = find_outer_ch tk Miz_semicolon in
  let tk2 = next_non_ws tk1 in
  let node = create_rec Permissive_assumption tk tk2 [] false in
  (node, tk2)

(** Definition-Item = Loci-Declaration | Permissive-Assumption | Auxiliary-Item . *)
let read_definition_item tk =
  let f (n, tk) = (create_rec Definition_item n.nt_first_tk n.nt_last_tk [n] true, tk) in
  match tk.tk_type with
  | Miz_let -> f (read_loci_declaration tk)
  | Miz_assume -> f (read_permissive_assumption tk)
  | Miz_given -> f (read_permissive_assumption tk)
  | _ -> f (read_auxiliary_item tk)

(* { Definition_item | Definition } *)
let rec read_def_list tk =
(*  let () = print_token tk in *)
(*  let () = print_endline "" in *)
  let f (node, tk) = (create_rec Definition node.nt_first_tk node.nt_last_tk [node] true, tk) in
  match tk.tk_type with
  | Miz_struct -> let (n, tk1) = f (read_structure_definition tk) in let (u, v) = read_def_list tk1 in (n :: u, v)
  | Miz_mode -> let (n, tk1) = f (read_mode_definition tk) in let (u, v) = read_def_list tk1 in (n :: u, v)
  | Miz_func -> let (n, tk1) = f (read_functor_definition tk) in let (u, v) = read_def_list tk1 in (n :: u, v)
  | Miz_pred -> let (n, tk1) = f (read_predicate_definition tk) in let (u, v) = read_def_list tk1 in (n :: u, v)
  | Miz_attr -> let (n, tk1) = f (read_attribute_definition tk) in let (u, v) = read_def_list tk1 in (n :: u, v)
  | Miz_canceled -> let (n, tk1) = f (read_canceled_definition tk) in let (u, v) = read_def_list tk1 in (n :: u, v)
  | Miz_redefine -> ([], tk)
  | Miz_end -> ([], tk)
  | _ -> let (n, tk1) = read_definition_item tk in let (u, v) = read_def_list tk1 in (n :: u, v)

let read_redefinition_block tk =
  (* let () = print_endline "redefinition" in *)
  let (ls, tk1) = read_def_list (next_non_ws tk) in
  let node = create_rec Redefinition_block tk tk1 ls true in
  (node, tk1)

let rec read_redefinition_block_list tk =
  match tk.tk_type with
  | Miz_redefine -> let (n, tk1) = read_redefinition_block tk in let (u, v) = read_redefinition_block_list tk1 in (n :: u, v)
  | _ -> ([], tk)

(** Definitional-Block = "definition" { ( Definition-Item | Definition ) } { Redefinition-Block } "end" . *)
let read_definitional_block tk =
  let (ls1, tk1) = read_def_list (next_non_ws tk) in
  let (ls2, tk2) = read_redefinition_block_list tk1 in
  if tk2.tk_type <> Miz_end then syntax_error_tk tk2 "Expected end"
  else
    let tk2 = next_non_ws tk2 in 
    let node = create_rec Definitional_block tk tk2 (ls1 @ls2) true in
    (node, tk2)

let read_definitional_item tk =
  let (n, tk1) = read_definitional_block tk in
  let () = if tk1.tk_type <> Miz_semicolon then syntax_error_tk tk1 "1 Expected ';'" in
  let tk2 = next_non_ws tk1 in
  let node = create_rec Definitional_item tk tk2 [n] true in
  (node, tk2)

let read_registration_item tk =
  let (n, tk1) = read_registration_block tk in
  let () = if tk1.tk_type <> Miz_semicolon then syntax_error_tk tk1 "2 Expected ';'" in
  let tk2 = next_non_ws tk1 in
  let node = create_rec Registration_item tk tk2 [n] true in
  (node, tk2)

let read_notation_item tk =
  let (n, tk1) = read_notation_block tk in
  let () = if tk1.tk_type <> Miz_semicolon then syntax_error_tk tk1 "3 Expected ';'" in
  let tk2 = next_non_ws tk1 in
  let node = create_rec Notation_item tk tk2 [n] true in
  (node, tk2)

let read_scheme_identifier tk =
  match tk.tk_type with
  | Miz_tk_ident -> (create_rec Scheme_identifier tk (next_non_ws tk) [] true, next_non_ws tk)
  | _ -> syntax_error_tk tk "Expected scheme identifier"

(** Scheme-Block = "scheme" Scheme-Identifier "{" Scheme-Parameters "}" ":" Scheme-Conclusion [ "provided" Scheme-Premise { "and" Scheme-Premise } ] Proof . *)
let read_scheme_block tk =
  let rec read_list tk_start tk_end =
    try
      let tk_and = find_first tk_start Miz_and tk_end in
      let n = create_rec Scheme_premise tk_start tk_and [] false in
      let u = read_list (next_non_ws tk_and) tk_end in
      n :: u
    with Not_found ->
      let n = create_rec Scheme_premise tk_start tk_end [] false in
      [n]
  in
  let () = if tk.tk_type <> Miz_scheme then syntax_error_tk tk "Expected scheme" in
  let (n1, tk1) = read_scheme_identifier (next_non_ws tk) in
  let () = if tk1.tk_type <> Miz_lc then syntax_error_tk tk1 "Expected '{'" in
  let tk2 = try find_matching_rc tk1 with Not_found -> syntax_error_tk tk1 "There is no matching '}'" in
  let n2 = create_rec Scheme_parameters tk1 tk2 [] false in
  let tk_colon = next_non_ws tk2 in
  let () = if tk_colon.tk_type <> Miz_colon then syntax_error_tk tk_colon "Expected ':'" in
  let tk_proof = try find_firsts_to_end tk_colon [Miz_proof; Miz_at_proof] with Not_found -> syntax_error_tk tk_colon "There is no proof for this scheme" in
  let ls =
    try
      let tk_provided = find_first tk_colon Miz_provided tk_proof in
      let n = create_rec Scheme_conclusion (next_non_ws tk_colon) tk_provided [] false in
      n :: (read_list (next_non_ws tk_provided) tk_proof)
    with Not_found -> [create_rec Scheme_conclusion (next_non_ws tk_colon) tk_proof [] false]
  in
  let tk_semi = try find_first_outer 1 (next_non_ws tk_proof) with Not_found -> syntax_error_tk tk_proof "There is no matching end" in
  let n_p = create_rec Proof tk_proof tk_semi [] false in
  let node = create_rec Scheme_block tk tk_semi ([n1; n2] @ ls @ [n_p]) true in
  (node, tk_semi)
  

(** Scheme-Item = Scheme-Block ";" . *)
let read_scheme_item tk =
  let (n, tk_semi) = read_scheme_block tk in
  let () = if tk_semi.tk_type <> Miz_semicolon then syntax_error_tk tk_semi "Expected ';'" in
  let tk_next = next_non_ws tk_semi in
  let node = create_rec Scheme_item tk tk_next [n] true in
  (node, tk_next)

let rec read_text_items tk =
  let f (n, tk1) =
    let node = create_rec Text_item tk tk1 [n] true in
    (node, tk1)
  in
  match tk.tk_type with
  | Miz_reserve -> let (n, tk1) = f (read_reservation tk) in let (u, v) = read_text_items tk1 in (n :: u, v)
  | Miz_definition -> let (n, tk1) = f (read_definitional_item tk) in let (u, v) = read_text_items tk1 in (n :: u, v)
  | Miz_registration -> let (n, tk1) = f (read_registration_item tk) in let (u, v) = read_text_items tk1 in (n :: u, v)
  | Miz_notation -> let (n, tk1) = f (read_notation_item tk) in let (u, v) = read_text_items tk1 in (n :: u, v)
  | Miz_theorem -> let (n, tk1) = f (read_text_item Theorem tk) in let (u, v) = read_text_items tk1 in (n :: u, v)
  | Miz_scheme -> let (n, tk1) = f (read_scheme_item tk) in let (u, v) = read_text_items tk1 in (n :: u, v)
  | Miz_set -> let (n, tk1) = f (read_auxiliary_item tk) in let (u, v) = read_text_items tk1 in (n :: u, v)
  | Miz_deffunc -> let (n, tk1) = f (read_auxiliary_item tk) in let (u, v) = read_text_items tk1 in (n :: u, v)
  | Miz_defpred -> let (n, tk1) = f (read_auxiliary_item tk) in let (u, v) = read_text_items tk1 in (n :: u, v)
  | Miz_canceled -> let (n, tk1) = f (read_canceled_theorem tk) in let (u, v) = read_text_items tk1 in (n :: u, v)
  | Miz_begin -> ([], tk) (* new Section *)
  | Miz_tk_eof   -> ([], tk) (* End of file *)
  | _ ->  let (n, tk1) = f (read_auxiliary_item tk) in let (u, v) = read_text_items tk1 in (n :: u, v) (* Auxiliary_item *)
  
(* Section { Section }             *)
(* Section = "begin" { Text_item } *)
let read_section tk =
  let () = if tk.tk_type <> Miz_begin then (syntax_error tk.tk_line tk.tk_pos "expected begin") in
  let tk1 = next_non_ws tk in
  let (ls, tk2) = read_text_items tk1 in
  let node = create_rec Section tk tk2 ls true in
  (node, tk2)

let rec read_sections tk =
  let (n, tk1) = read_section tk in
  match tk1.tk_type with
  | Miz_begin -> let (u, v) = read_sections tk1 in (n :: u, v)
  | Miz_tk_eof -> ([n], tk1)
  | _ -> syntax_error_tk tk1 "Expected begin or end of file"

(* Text_proper = Section { Section } *)
let read_text_proper tk =
  let (ls, tk1) = read_sections tk in
  let node = create_rec Text_proper tk tk1 ls true in
  (node, tk1)

(*  
let read_text_proper tk =
  let () = if tk.tk_type <> Miz_begin then (syntax_error tk.tk_line tk.tk_pos "expected begin") in
  let rec read_to_end tk =
    if tk.tk_type <> Miz_tk_eof then read_to_end (next_token tk) else tk in
  let tk1 = read_to_end tk in
  let node = create_rec Text_proper tk tk1 [] false in
  (node, tk1)
*)

let read_article first_tk =
  let tk = first_non_ws first_tk in
  let (env_node, tk1) = read_environ tk in
  let (text_node, tk2) = read_text_proper tk1 in
  let node = create_rec Article tk tk2 [env_node; text_node] true in
  (expand_tree node; node)

let tk_seq_to_xml buf tk_start tk_end =
  let rec f tk_opt =
    match tk_opt with
    Some tk1 when tk1 != tk_end -> (Buffer.add_string buf (tk_to_xml tk1); f tk1.tk_next)
    | None -> ()
    | _ -> ()
  in
    f (Some tk_start)

let tk_seq_to_xml_file out_s tk_start tk_end =
  let rec f tk_opt =
    match tk_opt with
    Some tk1 when tk1 != tk_end -> (output_string out_s (tk_to_xml tk1); f tk1.tk_next)
    | None -> ()
    | _ -> ()
  in
    f (Some tk_start)

let rec node_to_xml buf node =
  let rec add tk_start tk_end ls =
    match ls with
    | [] -> tk_seq_to_xml buf tk_start tk_end
    | h :: t -> (tk_seq_to_xml buf tk_start h.nt_first_tk; node_to_xml buf h; add h.nt_last_tk tk_end t)
  in
  let open_tag = open_ast_tag node.nt_type in
  let close_tag = close_ast_tag node.nt_type in
  let () = Buffer.add_string buf open_tag in
  let () = add node.nt_first_tk node.nt_last_tk node.nt_sublist in
  let () = Buffer.add_string buf close_tag in
  ()

let rec node_to_xml_file out_s node =
  let rec add tk_start tk_end ls =
    match ls with
    | [] -> tk_seq_to_xml_file out_s tk_start tk_end
    | h :: t -> (tk_seq_to_xml_file out_s tk_start h.nt_first_tk; node_to_xml_file out_s h; add h.nt_last_tk tk_end t)
  in
  let open_tag = open_ast_tag node.nt_type in
  let close_tag = close_ast_tag node.nt_type in
  let () = output_string out_s open_tag in
  let () = add node.nt_first_tk node.nt_last_tk node.nt_sublist in
  let () = output_string out_s close_tag in
  ()

let node_to_xml_filename xml_filename node =
  let out_s = open_out xml_filename in
  let () = node_to_xml_file out_s node in
  close_out out_s

let node_to_xml_string node =
  let buf = Buffer.create 16 in
  let () = node_to_xml buf node in
  Buffer.contents buf

let miz_file_to_ast_xml_file vct_map voc_dir miz_filename xml_filename =
  let lexer_start = Sys.time () in
  let tk = tokenize miz_filename vct_map voc_dir in
  let lexer_time = Sys.time () -. lexer_start in
  let () = print_endline ("tokens done in " ^ (string_of_float lexer_time)) in
  let parse_start = Sys.time () in
  let node = read_article tk in
  let list = collect_non_expanded node in
  let () = print_endline ("Number of non-expanded: " ^ (string_of_int (List.length list))) in
  let () = List.iter print_endline list in
  let parse_time = Sys.time () -. parse_start in
  let () = print_endline ("parse done in " ^ (string_of_float parse_time)) in
  (* let s = node_to_xml_string node in *)
  (* string_to_file s xml_filename *)
  node_to_xml_filename xml_filename node

module SetOfString = Set.Make(String)

let miz_file_to_ast_test vct_map voc_dir miz_filename =
  (* let lexer_start = Sys.time () in *)
  let tk = tokenize miz_filename vct_map voc_dir in
  (* let lexer_time = Sys.time () -. lexer_start in *)
  (* let () = print_endline ("tokens done in " ^ (string_of_float lexer_time)) in *)
  (* let parse_start = Sys.time () in *)
  let node = read_article tk in
  let list = collect_non_expanded node in
  (* let () = print_endline ("Number of non-expanded: " ^ (string_of_int (List.length list))) in *)
  (* let set = List.fold_left (fun set name -> SetOfString.add name set) SetOfString.empty list in *)
  (* set *)
  (* let () = print_endline ("Number of different non-expanded: " ^ (string_of_int (SetOfString.cardinal set))) in *)
  (* let () = List.iter print_endline (SetOfString.elements set) in () *)
  (* ignore (read_article tk) *)
  (* let parse_time = Sys.time () -. parse_start in *)
  (* let () = print_endline ("parse done in " ^ (string_of_float parse_time)) in *)
  list

let miz_file_to_ast_test1 vct_map voc_dir miz_filename =
  let tk = tokenize miz_filename vct_map voc_dir
  in read_article tk

let miz_file_to_ast vct_map voc_dir miz_filename =
  let tk = tokenize miz_filename vct_map voc_dir
  in read_article tk

let miz_file_to_list vct_map voc_dir miz_filename =
  let tk = tokenize miz_filename vct_map voc_dir in
  let node = read_article tk in
  collect_non_expanded node
